/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep.h: Interface to system-specific implementation routines.
 */

/*
 * This header declares the interface through which the platform-
 * independent routines exported by SIL call the actual implementations
 * specific to each platform.
 *
 * These routines are intended only for use by SIL itself; there is
 * normally no need to call them from outside.
 */

#ifndef SIL_SRC_SYSDEP_H
#define SIL_SRC_SYSDEP_H

struct InputEvent;
struct Matrix4f;
struct Vector2f;
struct Vector4f;
enum GraphicsPrimitiveID;
enum TextureFormat;

/* This is declared below in a typedef, but we also need it in some
 * functions declared earlier. */
struct SysTexture;

/*************************************************************************/
/********************** Macros for testing (part 1) **********************/
/*************************************************************************/

/*
 * For testing, we define separate versions of the sys_*() functions which
 * simulate the requisite functionality in a way which is reproducible and
 * which allows us to verify the results of each operation.  Since we
 * naturally cannot include two copies of an identically-named function
 * (much less expect the caller to know which one to call!), and since we
 * want to minimize the additional clutter in real sysdep implementations,
 * we write stubs here for every sysdep function which call either the real
 * or the test implementation depending on whether testing mode is enabled.
 * The actual stub functions and redirecting macros are defined at the
 * bottom of this file.
 *
 * In order for this functionality to work, any file which defines a
 * sys_*() function or which calls a specific implementation of such a
 * function must define either IN_SYSDEP (for real implementations) or
 * IN_SYSDEP_TEST (for test implementations).
 *
 * Note that we do not currently stub out the thread and mutex functions,
 * since their effects are directly testable (and since otherwise we'd
 * have to write a full-fledged scheduler and handle thread scheduling and
 * atomic operations ourselves).  Similarly, we don't stub out the file
 * and directory reading functions, instead relying on existing data (the
 * testdata directory at the top of the source tree) for testing.
 */

#if defined(INCLUDE_TESTS) && defined(IN_SYSDEP_TEST)
// Through the end of this section.

#define sys_input_init                  TEST_sys_input_init
#define sys_input_cleanup               TEST_sys_input_cleanup
#define sys_input_update                TEST_sys_input_update
#define sys_input_info                  TEST_sys_input_info
#define sys_input_grab                  TEST_sys_input_grab
#define sys_input_is_quit_requested     TEST_sys_input_is_quit_requested
#define sys_input_is_suspend_requested  TEST_sys_input_is_suspend_requested
#define sys_input_acknowledge_suspend_request TEST_sys_input_acknowledge_suspend_request
#define sys_input_joystick_button_mapping TEST_sys_input_joystick_button_mapping
#define sys_input_joystick_rumble       TEST_sys_input_joystick_rumble
#define sys_input_mouse_set_position    TEST_sys_input_mouse_set_position
#define sys_input_text_set_state        TEST_sys_input_text_set_state

#define sys_sound_init                  TEST_sys_sound_init
#define sys_sound_playback_rate         TEST_sys_sound_playback_rate
#define sys_sound_check_headphone_disconnect TEST_sys_sound_check_headphone_disconnect
#define sys_sound_acknowledge_headphone_disconnect TEST_sys_sound_acknowledge_headphone_disconnect
#define sys_sound_cleanup               TEST_sys_sound_cleanup
#define sys_sound_setdata               TEST_sys_sound_setdata
#define sys_sound_setvol                TEST_sys_sound_setvol
#define sys_sound_setpan                TEST_sys_sound_setpan
#define sys_sound_setfade               TEST_sys_sound_setfade
#define sys_sound_start                 TEST_sys_sound_start
#define sys_sound_stop                  TEST_sys_sound_stop
#define sys_sound_reset                 TEST_sys_sound_reset
#define sys_sound_status                TEST_sys_sound_status

#define sys_time_init                   TEST_sys_time_init
#define sys_time_now                    TEST_sys_time_now
#define sys_time_delay                  TEST_sys_time_delay

#define sys_userdata_get_data_path      TEST_sys_userdata_get_data_path
#define sys_userdata_perform            TEST_sys_userdata_perform

#define sys_display_error               TEST_sys_display_error
#define sys_get_language                TEST_sys_get_language
#define sys_open_file                   TEST_sys_open_file
#define sys_open_url                    TEST_sys_open_url
#define sys_reset_idle_timer            TEST_sys_reset_idle_timer

#define sys_debug_get_memory_stats      TEST_sys_debug_get_memory_stats

#endif  // INCLUDE_TESTS && IN_SYSDEP_TEST

/*************************************************************************/
/******************* Generic file and directory access *******************/
/*************************************************************************/

/**
 * SysFile:  Opaque file handle type, analogous to the FILE type defined in
 * the stdio library.
 */
typedef struct SysFile SysFile;

/**
 * SysDir: Opaque directory handle type, analogous to the DIR type defined
 * by dirent.h on POSIX systems.
 */
typedef struct SysDir SysDir;

/**
 * MAX_ASYNC_READS:  Maximum number of simultaneous asynchronous reads that
 * are supported by all systems (i.e., minimum number of simultaneous reads
 * that all systems must support).  Attempting to perform an asynchronous
 * read when this number of reads are outstanding may fail, depending on
 * the system.
 */
#define MAX_ASYNC_READS  100

/*--------------------------- File operations ---------------------------*/

/**
 * sys_file_open:  Open the given file from the filesystem.  Pathname
 * components are delimited with "/" characters and are treated
 * case-insensitively.  The meaning of multiple "/" characters in sequence
 * is undefined.
 *
 * If the native filesystem is case-sensitive and there are multiple files
 * which match the given pathname, one of the matching files will be
 * opened, but it is undefined which file is chosen -- _except_ that exact
 * matches for individual pathname components are preferred over inexact
 * matches, with earlier components in the pathname taking precedence over
 * later ones.
 *
 * For example, when requesting the path "abc/def/ghi" on a case-sensitive
 * filesystem, if the following three files exist in the filesystem:
 *     1) abc/Def/Ghi
 *     2) abc/DEF/GHI
 *     3) ABC/def/ghi
 * then either file 1 or file 2 will be chosen at random.  While file 3 has
 * more exact component matches ("def" and "ghi"), it loses to files 1 and
 * 2 on the first component and thus cannot be selected.
 *
 * If the pathname refers to a directory, the result is undefined; if the
 * function returns successfully for such a pathname, all operations on the
 * return handle except sys_file_close() are likewise undefined.
 *
 * [Parameters]
 *     path: Pathname of file to open.
 * [Return value]
 *     File handle, or NULL on error.
 */
extern SysFile *sys_file_open(const char *path);

/**
 * sys_file_dup:  Duplicate the given file handle.
 *
 * [Parameters]
 *     fh: File handle.
 * [Return value]
 *     Duplicated file handle, or NULL on error.
 */
extern SysFile *sys_file_dup(SysFile *fh);

/**
 * sys_file_close:  Close the given file.  If there are any outstanding
 * asynchronous read requests, they are automatically aborted as though
 * sys_file_abort_async() had been called on each such outstanding request.
 *
 * This function does nothing if fh == NULL.
 *
 * [Parameters]
 *     fh: File handle.
 */
extern void sys_file_close(SysFile *fh);

/**
 * sys_file_size:  Return the size in bytes of the given file.  This
 * function does not fail if passed a valid, open file handle.
 *
 * [Parameters]
 *     fh: File handle.
 * [Return value]
 *     Size of file, in bytes.
 */
extern int64_t sys_file_size(SysFile *fh);

/**
 * sys_file_seek:  Set the position for synchronous file reads.  This
 * function does not fail if passed a valid, open file handle and a valid
 * FILE_SEEK_* constant for "how".
 *
 * [Parameters]
 *     fh: File handle.
 *     pos: New synchronous read position, in bytes.
 *     how: Method for setting position (FILE_SEEK_*).
 * [Return value]
 *     True on success, false on error.
 */
extern int sys_file_seek(SysFile *fh, int64_t pos, int how);
enum {
    FILE_SEEK_SET = 0,  // Set the read position to the given value.
    FILE_SEEK_CUR,      // Add the given value to the current read position.
    FILE_SEEK_END,      // Set the position to the given value + the file size.
};

/**
 * sys_file_tell:  Return the current position for synchronous file reads.
 * This function does not fail if passed a valid, open file handle.
 *
 * [Parameters]
 *     fh: File handle.
 * [Return value]
 *     Current synchronous read position, in bytes.
 */
extern int64_t sys_file_tell(SysFile *fh);

/**
 * sys_file_read:  Read data synchronously from the given file, starting
 * at the current synchronous read position.  If any asynchronous reads
 * are in progress on the file, this function may wait until those reads
 * have completed.
 *
 * [Parameters]
 *     fh: File handle.
 *     buf: Read buffer.
 *     len: Number of bytes to read.
 * [Return value]
 *     Number of bytes read, or a negative value on error.  End-of-file is
 *     is not considered an error.
 */
extern int32_t sys_file_read(SysFile *fh, void *buf, int32_t len);

/**
 * sys_file_read_at:  Read data synchronously from the given file, starting
 * at the given position.  If any asynchronous reads are in progress on the
 * file, this function may wait until those reads have completed.
 *
 * This function does not affect the current synchronous read position used
 * by sys_file_seek() and sys_file_read().
 *
 * [Parameters]
 *     fh: File handle.
 *     buf: Read buffer.
 *     len: Number of bytes to read.
 *     filepos: Starting position for read, in bytes from the beginning of
 *         the file.
 * [Return value]
 *     Number of bytes read, or a negative value on error.  End-of-file is
 *     not considered an error.
 */
extern int32_t sys_file_read_at(SysFile *fh, void *buf, int32_t len, int64_t filepos);

/**
 * sys_file_read_async:  Start an asynchronous read from the given file,
 * and return immediately.  The read buffer may not be accessed until
 * completion of the read has been confirmed by calling sys_file_wait_async().
 *
 * The read request ID returned by this function should be considered
 * thread-local, and must not be used from other threads.
 *
 * Asynchronous reads do not affect the current synchronous read position.
 *
 * [Parameters]
 *     fh: File handle.
 *     buf: Read buffer.
 *     len: Number of bytes to read.
 *     filepos: Starting position for read, in bytes.
 * [Return value]
 *     Read request ID (nonzero), or zero on error.
 */
extern int sys_file_read_async(SysFile *fh, void *buf, int32_t len, int64_t filepos);

/**
 * sys_file_poll_async:  Check whether the given asynchronous read has
 * has completed.
 *
 * [Parameters]
 *     request: Read request ID.
 * [Return value]
 *     True if the read has completed (whether successfully or not) or if
 *     the request ID is invalid; false if the read is in progress.
 */
extern int sys_file_poll_async(int request);

/**
 * sys_file_wait_async:  Wait (if necessary) for the given asynchronous
 * read to complete, and return the result of the read.  The result is the
 * same as would have been returned by sys_file_read() with the same
 * parameters from the same read position.
 *
 * This function will never block for a given read request if
 * sys_file_poll_async() returns true for that request.
 *
 * [Parameters]
 *     request: Read request ID.
 * [Return value]
 *     Number of bytes read, or a negative value on error.  End-of-file is
 *     not considered an error.
 */
extern int32_t sys_file_wait_async(int request);

/**
 * sys_file_abort_async:  Abort the given asynchronous read.  Even if this
 * function succeeds, the buffer must not be freed or reused until the
 * request's completion has been confirmed with sys_file_wait_async().
 * After calling this function, the request's result will always be an
 * error, but the state of the buffer is undefined.
 *
 * Some systems may not be able to immediately abort an asynchronous read
 * request, so a successful return from this function does _not_ guarantee
 * that a subsequent sys_file_wait_async() call on the same request will
 * return immediately.
 *
 * [Parameters]
 *     request: Read request ID.
 * [Return value]
 *     True on success, false on error.
 */
extern int sys_file_abort_async(int request);

/*------------------------ Directory operations -------------------------*/

/**
 * sys_dir_open:  Open the given directory from the filesystem.  Pathname
 * components are delimited with "/" characters and are treated
 * case-insensitively, following the same rules as for sys_file_open().
 *
 * [Parameters]
 *     path: Pathname of directory to open.
 * [Return value]
 *     Directory handle, or NULL on error.
 */
extern SysDir *sys_dir_open(const char *path);

/**
 * sys_dir_read:  Read the next file entry from the given directory and
 * return its filename (not including the directory pathname).  Files are
 * not guaranteed to be returned in any particular order, but each
 * distinct file in the directory is guaranteed to be returned only once.
 * Subdirectories and other non-file entries are _not_ returned.
 *
 * Note that the returned string is stored in a static buffer which will
 * be overwritten by the next call to sys_dir_read() for this handle and
 * is only valid until sys_dir_close() is called for the handle.
 *
 * [Parameters]
 *     dir: Directory handle.
 * [Return value]
 *     Fiilename of next file, or NULL if all file entries have been read.
 */
extern const char *sys_dir_read(SysDir *dir);

/**
 * sys_dir_close:  Close the given directory.  This function does nothing
 * if passed a NULL value.
 *
 * [Parameters]
 *     dir: Directory handle.
 */
extern void sys_dir_close(SysDir *dir);

/*************************************************************************/
/************************* Framebuffer handling **************************/
/*************************************************************************/

/**
 * SysFramebuffer:  Opaque type for system-specific framebuffer data.
 */
typedef struct SysFramebuffer SysFramebuffer;

/*-----------------------------------------------------------------------*/

/**
 * sys_framebuffer_supported:  Return whether the system supports
 * offscreen framebuffers.  Implements the high-level function
 * graphics_framebuffer_supported().
 *
 * [Return value]
 *     True if offscreen framebuffers are supported, false if not.
 */
extern int sys_framebuffer_supported(void);

/**
 * sys_framebuffer_create:  Create a new framebuffer of the given size.
 * If the system does not support framebuffers, this function will always
 * return NULL.
 *
 * [Parameters]
 *     width, height: Framebuffer size, in pixels (guaranteed >=1).
 * [Return value]
 *     New framebuffer, or NULL on error.
 */
extern SysFramebuffer *sys_framebuffer_create(int width, int height);

/**
 * sys_framebuffer_destroy:  Destroy the given framebuffer.
 *
 * [Parameters]
 *     framebuffer: Framebuffer to destroy (guaranteed non-NULL).
 */
extern void sys_framebuffer_destroy(SysFramebuffer *framebuffer);

/**
 * sys_framebuffer_bind:  Bind the given framebuffer to the current
 * rendering target.  If framebuffer == NULL, cancel any framebuffer
 * binding so that the display is bound to the rendering target.
 *
 * [Parameters]
 *     framebuffer: Framebuffer to bind, or NULL to cancel binding.
 */
extern void sys_framebuffer_bind(SysFramebuffer *framebuffer);

/**
 * sys_framebuffer_apply:  Apply the given framebuffer as a texture for
 * rendering.
 *
 * [Parameters]
 *     framebuffer: Framebuffer to apply (guaranteed non-NULL).
 */
extern void sys_framebuffer_apply(SysFramebuffer *framebuffer);

/**
 * sys_framebuffer_set_antialias:  Set the given framebuffer's antialiasing
 * flag.
 *
 * [Parameters]
 *     framebuffer: Framebuffer to modify.
 *     on: True to enable antialiasing, false to disable.
 */
extern void sys_framebuffer_set_antialias(SysFramebuffer *framebuffer, int on);

/**
 * sys_framebuffer_discard_data:  Discard all pixel data in the given
 * framebuffer.  Implements the high-level function
 * graphics_discard_framebuffer_data().
 *
 * [Parameters]
 *     framebuffer: Framebuffer whose contents may be discarded (guaranteed
 *         non-NULL).
 */
extern void sys_framebuffer_discard_data(SysFramebuffer *framebuffer);

/*************************************************************************/
/***************** Graphics and rendering functionality ******************/
/*************************************************************************/

/*
 * Many of these functions are simply implementation functions for the
 * high-level graphics interface in graphics.h; however, these functions
 * may assume that their parameters have been checked for validity, unless
 * otherwise specified in the function documentation.
 */

/*-------------- Graphics-related constants and structures --------------*/

/**
 * SysGraphicsSizeEntry:  Simple structure to hold a display size entry.
 */
typedef struct SysGraphicsSizeEntry SysGraphicsSizeEntry;
struct SysGraphicsSizeEntry {
    int width, height;
};

/**
 * SysGraphicsInfo:  Structure describing the system's graphics capabilities.
 */
typedef struct SysGraphicsInfo SysGraphicsInfo;
struct SysGraphicsInfo {
    /* True if a windowed mode is supported, as for
     * graphics_has_windowed_mode(). */
    uint8_t has_windowed_mode;

    /* Number and array of supported display sizes.  At least one valid
     * size must be returned from sys_graphics_init().  The pointer is
     * declared const to enforce read-onlyness to the caller, but the
     * implementation is free to cast it away if needed. */
    int num_sizes;
    const SysGraphicsSizeEntry *sizes;
};

/**
 * SysGraphicsParam:  Constants for standard rendering parameters, used
 * with sys_graphics_{set,get}_*_param().  These values are all negative,
 * so that positive values can be used for user-defined parameters.
 * (Note that user-defined parameters are not yet implemented.)
 */
enum SysGraphicsParam {
    /* Coordinate transformation matrices. */
    SYS_GRAPHICS_PARAM_PROJECTION_MATRIX     = -1,
    SYS_GRAPHICS_PARAM_VIEW_MATRIX           = -2,
    SYS_GRAPHICS_PARAM_MODEL_MATRIX          = -3,

    /* Rendering parameters. */
    SYS_GRAPHICS_PARAM_ALPHA_TEST            = -101, // boolean
    SYS_GRAPHICS_PARAM_ALPHA_TEST_COMPARISON = -102, // int (GRAPHICS_COMPARISON_*)
    SYS_GRAPHICS_PARAM_ALPHA_REFERENCE       = -103, // float
    SYS_GRAPHICS_PARAM_BLEND_COLOR           = -104, // Vector4f
    SYS_GRAPHICS_PARAM_CLIP                  = -105, // boolean
    SYS_GRAPHICS_PARAM_COLOR_WRITE           = -106, // boolean[3] (bit0=R, bit1=G, bit2=B)
    SYS_GRAPHICS_PARAM_DEPTH_TEST            = -107, // boolean test enabled?
    SYS_GRAPHICS_PARAM_DEPTH_TEST_COMPARISON = -108, // int (GRAPHICS_COMPARISON_*)
    SYS_GRAPHICS_PARAM_DEPTH_WRITE           = -109, // boolean
    SYS_GRAPHICS_PARAM_FACE_CULL             = -110, // boolean
    SYS_GRAPHICS_PARAM_FACE_CULL_CW          = -111, // boolean (true=CW, false=CCW)
    SYS_GRAPHICS_PARAM_FIXED_COLOR           = -112, // Vector4f
    SYS_GRAPHICS_PARAM_FOG                   = -113, // boolean
    SYS_GRAPHICS_PARAM_FOG_START             = -114, // float
    SYS_GRAPHICS_PARAM_FOG_END               = -115, // float
    SYS_GRAPHICS_PARAM_FOG_COLOR             = -116, // Vector4f
    SYS_GRAPHICS_PARAM_LINE_WIDTH            = -117, // float
    SYS_GRAPHICS_PARAM_POINT_SIZE            = -118, // float
    SYS_GRAPHICS_PARAM_STENCIL_TEST          = -119, // boolean
    /* These three values are always set as a group, STENCIL_MASK last. */
    SYS_GRAPHICS_PARAM_STENCIL_COMPARISON    = -120, // int (GRAPHICS_COMPARISON_*)
    SYS_GRAPHICS_PARAM_STENCIL_REFERENCE     = -121, // unsigned int
    SYS_GRAPHICS_PARAM_STENCIL_MASK          = -122, // unsigned int
    /* These three values are always set as a group, STENCIL_OP_DPASS last. */
    SYS_GRAPHICS_PARAM_STENCIL_OP_SFAIL      = -123, // int (GRAPHICS_STENCIL_*)
    SYS_GRAPHICS_PARAM_STENCIL_OP_DFAIL      = -124, // int (GRAPHICS_STENCIL_*)
    SYS_GRAPHICS_PARAM_STENCIL_OP_DPASS      = -125, // int (GRAPHICS_STENCIL_*)

    /* Texture mapping parameters. */
    SYS_GRAPHICS_PARAM_TEXTURE_OFFSET        = -201, // Vector2f
};
typedef enum SysGraphicsParam SysGraphicsParam;

/**
 * SysPrimitive:  Opaque structure representing a graphics primitive and
 * associated vertex data, used with the sys_graphics_*_primitive() functions.
 */
typedef struct SysPrimitive SysPrimitive;

/*---------------------- Basic graphics operations ----------------------*/

/**
 * sys_graphics_init:  Initialize the graphics/rendering functionality,
 * and return a structure describing the system's graphics capabilities.
 *
 * [Return value]
 *     Pointer to graphics capability structure, or NULL on error.
 */
extern const SysGraphicsInfo *sys_graphics_init(void);

/**
 * sys_graphics_cleanup:  Shut down the graphics/rendering functionality,
 * close the display (if open), and free all related resources.
 */
extern void sys_graphics_cleanup(void);

/**
 * sys_graphics_device_width, sys_graphics_device_height:  Return the
 * current width or height of the display device.  These implement the
 * high-level functions graphics_device_width() and graphics_device_height().
 *
 * [Return value]
 *     Display device width or height, in pixels.
 */
extern int sys_graphics_device_width(void);
extern int sys_graphics_device_height(void);

/**
 * sys_graphics_set_display_attr:  Set a system-specific display attribute.
 * Implements the high-level function graphics_set_display_mode().
 *
 * [Parameters]
 *     name: Attribute name (guaranteed non-NULL).
 *     args: Attribute value(s).
 * [Return value]
 *     True if the display attribute was successfully set, false if the
 *     attribute is unknown or an error occurred.
 */
extern int sys_graphics_set_display_attr(const char *name, va_list args);

/**
 * sys_graphics_set_display_mode:  Set a new display mode.  Implements
 * the high-level function graphics_set_display_mode().
 *
 * [Parameters]
 *     width, height: Display size, in pixels.
 * [Return value]
 *     True if the display mode was successfully set, false on error.
 */
extern int sys_graphics_set_display_mode(int width, int height);

/**
 * sys_graphics_renderer_info:  Return a string describing the low-level
 * rendering backend.  Implements the high-level function
 * graphics_renderer_info().
 *
 * [Return value]
 *     Renderer information string.
 */
extern const char *sys_graphics_renderer_info(void);

/**
 * sys_graphics_display_is_window:  Return whether the current display
 * mode is a windowed mode.  Implements the high-level function
 * graphics_display_is_window().
 *
 * [Return value]
 *     True if the display mode is a windowed mode, false if the display
 *     mode is a full-screen mode or if no display mode is set.
 */
extern int sys_graphics_display_is_window(void);

/**
 * sys_graphics_set_window_title:  Set the title to be displayed in the
 * window title bar or any equivalent system facility.  Implements the
 * high-level function graphics_set_window_title().
 *
 * [Parameters]
 *     title: Title string (guaranteed non-NULL).
 */
extern void sys_graphics_set_window_title(const char *title);

/**
 * sys_graphics_set_window_icon:  Set the icon to be associated with the
 * window.  Implements the high-level function graphics_set_window_icon().
 *
 * [Parameters]
 *     texture: Texture to use as icon image (guaranteed non-NULL and
 *         unlocked).
 */
extern void sys_graphics_set_window_icon(struct SysTexture *texture);

/**
 * sys_graphics_show_mouse_pointer:  Set whether to display a system-level
 * visible mouse pointer on top of the display.  Implements the high-level
 * function graphics_show_mouse_pointer().
 *
 * [Parameters]
 *     on: True to display the pointer, false to hide it.
 */
extern void sys_graphics_show_mouse_pointer(int on);

/**
 * sys_graphics_get_mouse_pointer_state:  Return whether the system-level
 * mouse pointer display is enabled.  Implements the high-level function
 * graphics_get_mouse_pointer_state().
 *
 * [Return value]
 *     True if the pointer is displayed, false if not.
 */
extern int sys_graphics_get_mouse_pointer_state(void);

/**
 * sys_graphics_get_frame_period:  Return the nominal frame period of the
 * display in seconds, based on the current display mode and attributes.
 * The frame period is defined as the mean time between successive calls
 * to sys_graphics_finish_frame() under the assumptions that (1) the
 * intervening calls to sys_graphics_start_frame() block until the display
 * hardware is ready to accept a new frame and (2) all other graphics
 * calls complete immediately.
 *
 * If the frame period is unknown, or if the display does not have a fixed
 * frame period (for example, because vertical sync has been disabled), the
 * function returns a period of zero (numerator == 0, denominator != 0).
 *
 * The behavior of this function is undefined if called before a display
 * mode has been successfully set (including after a failed call to
 * sys_graphics_set_display_mode()).
 *
 * [Parameters]
 *     numerator_ret: Pointer to variable to receive the numerator of the
 *         frame period (guaranteed non-NULL).
 *     denominator_ret: Pointer to variable to receive the denominator of
 *         the frame period (guaranteed non-NULL).
 */
extern void sys_graphics_get_frame_period(int *numerator_ret,
                                          int *denominator_ret);

/**
 * sys_graphics_has_focus:  Return whether this program has the input
 * focus.  Implements the high-level function graphics_has_focus().
 *
 * [Return value]
 *     True if this program has the input focus, false if not.
 */
extern int sys_graphics_has_focus(void);

/**
 * sys_graphics_start_frame:  Prepare for rendering a new frame, and return
 * the current width and height of the display.  (These values will be
 * treated as constant until the next call to sys_graphics_finish_frame().)
 *
 * This function may freely change the framebuffer binding; the high-level
 * function graphics_start_frame() will call sys_framebuffer_bind(NULL)
 * after this function returns.
 *
 * [Parameters]
 *     width_ret, height_ret: Pointers to variables to receive the current
 *         display size, in pixels (guaranteed non-NULL).
 */
extern void sys_graphics_start_frame(int *width_ret, int *height_ret);

/**
 * sys_graphics_finish_frame:  Finish rendering the current frame, and
 * perform any operations necessary to present the rendered framebuffer
 * to the user (such as flipping buffers for a double-buffered display).
 * This function should _not_ wait for background hardware rendering to
 * complete or wait for the next vertical blank unless required by the
 * system.
 *
 * This function may freely change the framebuffer binding.
 */
extern void sys_graphics_finish_frame(void);

/**
 * sys_graphics_sync:  Wait for any background rendering operations to
 * complete, and optionally flush any resources which are pending deletion.
 *
 * [Parameters]
 *     flush: True to release resources pending deletion.
 */
extern void sys_graphics_sync(int flush);

/**
 * sys_graphics_clear:  Clear the display.  Implements the high-level
 * function graphics_clear().
 *
 * [Parameters]
 *     r, g, b: Clear color.
 */
extern void sys_graphics_clear(float r, float g, float b);

/**
 * sys_graphics_read_pixels:  Read pixels from the display.  Implements
 * the high-level function graphics_read_pixels().
 *
 * [Parameters]
 *     x, y: Pixel coordinates of upper-left corner of region to read
 *         (0,0 is the bottom-left corner of the display); both are
 *         guaranteed to be nonnegative.
 *     w, h: Size of region to read, in pixels; both are guaranteed to
 *         be positive.
 *     stride: Line stride of buffer, in pixels; guaranteed to be >= w.
 *     buffer: Buffer into which pixel data will be stored; guaranteed to
 *         be non-NULL and have room for at least w*h*4 bytes.
 * [Return value]
 *     True on success, false on error.
 */
extern int sys_graphics_read_pixels(int x, int y, int w, int h, int stride,
                                    void *buffer);

/*---------------------- Render state manipulation ----------------------*/

/*
 * Note that the high-level graphics routines do not attempt to check for
 * redundant state changes (i.e., state change operations that set a state
 * parameter to its current value or that are subsequently changed again
 * without being used in a draw operation), so sysdep implementations
 * should add such checks if they would be significantly less expensive
 * than just performing the redundant change.
 */

/**
 * sys_graphics_set_viewport:  Set the render viewport position and size.
 * Implements the high-level function graphics_set_viewport().
 *
 * [Parameters]
 *     left: Leftmost X coordinate of the viewport, in pixels
 *         (guaranteed >=0).
 *     bottom: Bottommost Y coordinate of the viewport, in pixels
 *         (guaranteed >=0).
 *     width: Width of the viewport, in pixels (guaranteed >=1).
 *     height: Height of the viewport, in pixels (guaranteed >=1).
 */
extern void sys_graphics_set_viewport(int left, int bottom,
                                      int width, int height);

/**
 * sys_graphics_set_clip_region:  Set the render clipping region and size.
 * Implements the high-level function graphics_set_clip_region(); note
 * that the case of width==0 && height==0 (reset clipping region) is
 * translated to proper values by graphics_set_clip_region().
 *
 * [Parameters]
 *     left: Leftmost X coordinate of the clipping region, in pixels
 *         (guaranteed >=0).
 *     bottom: Bottommost Y coordinate of the clipping region, in pixels
 *         (guaranteed >=0).
 *     width: Width of the clipping region, in pixels (guaranteed >=0).
 *     height: Height of the clipping region, in pixels (guaranteed >=0).
 */
extern void sys_graphics_set_clip_region(int left, int bottom,
                                         int width, int height);

/**
 * sys_graphics_set_depth_range:  Set the depth range mapping.
 * Implements the high-level function graphics_set_depth_range(); the
 * caller guarantees that both parameters are within [-1,+1].
 *
 * [Parameters]
 *     near: Window coordinate of the near clipping plane.
 *     far: Window coordinate of the far clipping plane.
 */
extern void sys_graphics_set_depth_range(float near, float far);

/**
 * sys_graphics_set_blend:  Set the pixel blending operation and pixel factors.
 *
 * [Parameters]
 *     operation: Blending operation (GRAPHICS_BLEND_ADD, etc.).
 *     src_factor: Source (new) pixel blending factor (GRAPHICS_BLEND_*).
 *     dest_factor: Destination (old) pixel blending factor (GRAPHICS_BLEND_*).
 */
extern void sys_graphics_set_blend(int operation, int src_factor, int dest_factor);

/**
 * sys_graphics_set_*_param:  Set a rendering parameter.  The caller will
 * ensure the value and parameter types match (booleans are passed as ints).
 *
 * [Parameters]
 *     id: Parameter ID.
 *     value: Value to set.  For booleans, a nonzero value is true and
 *         zero is false.
 */
extern void sys_graphics_set_int_param(int id, int value);
extern void sys_graphics_set_float_param(int id, float value);
extern void sys_graphics_set_vec2_param(int id, const struct Vector2f *value);
extern void sys_graphics_set_vec4_param(int id, const struct Vector4f *value);
extern void sys_graphics_set_matrix_param(int id, const struct Matrix4f *value);

/**
 * sys_graphics_get_matrix_param:  Retrieve the value of a matrix-type
 * rendering parameter.
 *
 * [Parameters]
 *     id: Parameter ID.
 *     value_ret: Pointer to variable into which to store parameter value.
 */
extern void sys_graphics_get_matrix_param(int id, struct Matrix4f *value_ret);

/*------------------ Primitive creation and rendering -------------------*/

/**
 * sys_graphics_create_primitive:  Create a new primitive given the
 * primitive type and vertex list.  If immediate == 0, the vertex data
 * buffer will be copied, so it may be reused or destroyed after this
 * function returns.
 *
 * The "format" parameter describes the format of the vertex data, and
 * points to an array of uint32_t values generated by the
 * GRAPHICS_VERTEX_FORMAT() macro, terminated by a value of zero.  The
 * caller guarantees that at least a vertex position will be included.
 *
 * Optionally, a list of indices can be provided in index_data.  The size
 * of each index is specified by index_size as 1, 2, or 4, representing
 * unsigned 8-bit, 16-bit, or 32-bit integers respectively.
 *
 * If the "immediate" parameter is true, then the function may assume that
 * the primitive will be drawn and destroyed immediately without any other
 * intervening primitive calls.  In particular, the function may choose to
 * render the primitive immediately and ignore the following
 * sys_graphics_draw_primitive() call.  The passed-in buffers may also be
 * reused, so the caller must ensure they remain valid until
 * sys_graphics_draw_primitive() returns.
 *
 * [Parameters]
 *     type: Primitive type (GRAPHICS_PRIMITIVE_*, guaranteed valid).
 *     data: Vertex data pointer (guaranteed non-NULL).
 *     format: Vertex data format description (guaranteed non-NULL).
 *     size: Size of a single vertex, in bytes (guaranteed >0).
 *     count: Number of vertices (guaranteed >0).
 *     index_data: Index data pointer, or NULL if indices are not used.
 *     index_size: Size of an index value, in bytes (guaranteed valid if
 *         index_data != NULL).
 *     index_count: Number of indices (guaranteed >0 if index_data != NULL).
 *     immediate: Optimization hint: if true, this primitive will be drawn
 *         and destroyed immediately.
 * [Return value]
 *     New primitive, or NULL on error.
 */
extern SysPrimitive *sys_graphics_create_primitive(
    enum GraphicsPrimitiveID type, const void *data, const uint32_t *format,
    int size, int count, const void *index_data,
    int index_size, int index_count, int immediate);

/**
 * sys_graphics_draw_primitive:  Draw a previously-created primitive.
 *
 * [Parameters]
 *     primitive: Primitive to draw (guaranteed non-NULL).
 */
extern void sys_graphics_draw_primitive(SysPrimitive *primitive);

/**
 * sys_graphics_destroy_primitive:  Destroy a primitive.
 *
 * [Parameters]
 *     primitive: Primitive to destroy (guaranteed non-NULL).
 */
extern void sys_graphics_destroy_primitive(SysPrimitive *primitive);

/*************************************************************************/
/************** Text rendering using a system-provided font **************/
/*************************************************************************/

/**
 * SysFont:  Opaque structure representing a system-provided font, used
 * with the sys_sysfont_*() functions.
 */
typedef struct SysFont SysFont;

/*-----------------------------------------------------------------------*/

/**
 * sys_sysfont_create:  Create a new reference to a system-provided font.
 *
 * The font name passed to this function can be interpreted as appropriate
 * for the system, except that the empty string should select a default
 * font (which must support at least the ASCII character set, U+0020
 * through U+007E).  If the requested font is not available, this function
 * should return NULL rather than attempting to find a "similar" font
 * (however, requests for the default font should always succeed unless
 * the system does not support fonts at all).
 *
 * The size parameter indicates the character size at which the font is
 * intended to be used, but operations on the font may use any size.
 * If the system is unable to scale the font to arbitrary sizes, render
 * operations should return an appropriate scale factor in the "scale_ret"
 * parameter to sys_sysfont_render(), which will cause the high-level code
 * to expand the rendered text to the appropriate size.
 *
 * [Parameters]
 *     name: System-dependent font name, or the empty string for the
 *         default font.
 *     size: Desired font size, in pixels.
 *     mem_flags: Memory allocation flags (MEM_ALLOC_*).
 * [Return value]
 *     SysFont object for the font, or NULL on error.
 */
extern SysFont *sys_sysfont_create(const char *name, float size, int mem_flags);

/**
 * sys_sysfont_destroy:  Free resources used by a SysFont object.
 *
 * [Parameters]
 *     font: SysFont object to destroy (guaranteed non-NULL).
 */
extern void sys_sysfont_destroy(SysFont *font);

/**
 * sys_sysfont_native_size:  Return the "native" size of the given font,
 * i.e., the size at which the font is designed to be rendered.  In the
 * case of a vector or similar font which can be rendered equally well at
 * any size, the "native" size is zero.
 *
 * [Parameters]
 *     font: Font reference (guaranteed non-NULL).
 * [Return value]
 *     Native size of font, in pixels, or zero if none.
 */
extern int sys_sysfont_native_size(SysFont *font);

/**
 * sys_sysfont_get_metrics:  Return font size metrics for text drawn at
 * the given font size.  The values returned are defined as for the
 * high-level functions font_height(), font_baseline(), font_ascent(),
 * and font_descent().
 *
 * [Parameters]
 *     font: Font reference (guaranteed non-NULL).
 *     size: Font size, in pixels (guaranteed >0).
 *     height_ret: Pointer to variable to receive the line height, in pixels
 *         (guaranteed non-NULL).
 *     baseline_ret: Pointer to variable to receive the baseline offset, in
 *         pixels (guaranteed non-NULL).
 *     ascent_ret: Pointer to variable to receive the font ascent, in pixels
 *         (guaranteed non-NULL).
 *     descent_ret: Pointer to variable to receive the font descent, in pixels
 *         (guaranteed non-NULL).
 */
extern void sys_sysfont_get_metrics(
    SysFont *font, float size, float *height_ret, float *baseline_ret,
    float *ascent_ret, float *descent_ret);

/**
 * sys_sysfont_char_advance:  Return the horizontal advance for the given
 * character at the given font size.  Implements the high-level function
 * font_char_advance() for system-provided fonts.
 *
 * [Parameters]
 *     font: Font reference (guaranteed non-NULL).
 *     ch: Unicode character (codepoint).
 *     size: Font size, in pixels (guaranteed >0).
 * [Return value]
 *     Horizontal advance, in pixels.
 */
extern float sys_sysfont_char_advance(SysFont *font, int32_t ch, float size);

/**
 * sys_sysfont_text_advance:  Return the horizontal advance for the given
 * text string at the given font size.  Implements the high-level function
 * font_text_advance() for system-provided fonts.
 *
 * [Parameters]
 *     font: Font reference (guaranteed non-NULL).
 *     str: UTF-8 text string (guaranteed non-NULL).
 *     size: Font size, in pixels (guaranteed >0).
 * [Return value]
 *     Horizontal advance, in pixels.
 */
extern float sys_sysfont_text_advance(SysFont *font, const char *str, float size);

/**
 * sys_sysfont_get_text_bounds:  Return the horizontal bounds of the given
 * text string as it would be rendered at the given font size.
 *
 * [Parameters]
 *     font: Font reference (guaranteed non-NULL).
 *     str: UTF-8 text string (guaranteed non-NULL).
 *     size: Font size, in pixels (guaranteed >0).
 *     left_ret: Pointer to variable to receive the distance from the origin
 *         to the left edge of the text, in pixels (guaranteed non-NULL).
 *     right_ret: Pointer to variable to receive the distance from the origin
 *         to the right edge of the text, in pixels (guaranteed non-NULL).
 */
extern void sys_sysfont_get_text_bounds(
    SysFont *font, const char *str, float size, float *left_ret,
    float *right_ret);

/**
 * sys_sysfont_render:  Render the given text to a texture.
 *
 * [Parameters]
 *     font: Font reference (guaranteed non-NULL).
 *     str: UTF-8 text string (guaranteed non-NULL).
 *     size: Font size, in pixels (guaranteed >0).
 *     origin_x_ret: Pointer to variable to receive the texture X coordinate
 *         of the text origin, in pixels (guaranteed non-NULL).
 *     origin_y_ret: Pointer to variable to receive the texture Y coordinate
 *         of the text origin, in pixels (guaranteed non-NULL).
 *     advance_ret: Pointer to variable to receive the horizontal advance
 *         from the text origin, in pixels (guaranteed non-NULL).
 *     scale_ret: Pointer to variable to receive the scale factor at which
 *         the texture should be drawn (guaranteed non-NULL).
 * [Return value]
 *     Newly created texture, or NULL on error.
 */
extern struct SysTexture *sys_sysfont_render(
    SysFont *text, const char *str, float size, float *origin_x_ret,
    float *origin_y_ret, float *advance_ret, float *scale_ret);

/*************************************************************************/
/************************* Input device handling *************************/
/*************************************************************************/

/**
 * SysInputJoyInfo:  Information about a single joystick.
 */
typedef struct SysInputJoystick SysInputJoystick;
struct SysInputJoystick {
    /* Is the joystick at this index connected? */
    uint8_t connected;

    /* Does this joystick support rumble (force feedback)? */
    uint8_t can_rumble;

    /* Number of buttons on the joystick (counting from zero, so that the
     * highest valid butten number is num_buttons-1). */
    int num_buttons;

    /* Number of sticks on the joystick (counting from zero). */
    int num_sticks;
};

/**
 * SysInputInfo:  Information about available input devices on the system.
 */
typedef struct SysInputInfo SysInputInfo;
struct SysInputInfo {
    /* Is a joystick available? */
    uint8_t has_joystick;

    /* Number of joystick devices available (whether a joystick is actually
     * connected or not). */
    int num_joysticks;

    /* Array of num_joysticks SysInputJoystick structures describing the
     * individual joystick devices.  The array pointed to by this field
     * is guaranteed to remain valid until the next sys_input_info() or
     * sys_input_cleanup() call. */
    const SysInputJoystick *joysticks;

    /*------------------------------*/

    /* Is a keyboard, keypad, or similar key-based input device available? */
    uint8_t has_keyboard;

    /* Is the key-based input device a "full" keyboard suitable for text
     * entry? */
    uint8_t keyboard_is_full;

    /*------------------------------*/

    /* Is a mouse available? */
    uint8_t has_mouse;

    /*------------------------------*/

    /* Is a generic text entry interface (such as a software keyboard)
     * available?  Systems with true keyboards are _not_ required to
     * implement this interface, though it is recommended to ensure proper
     * translation of locale-specific keys and key combinations.  If a
     * keyboard is available and this field is set to zero, a default
     * translation table will be used to provide text entry services via
     * the keyboard. */
    uint8_t has_text;

    /* Does the text entry functionality use its own display interface? */
    uint8_t text_uses_custom_interface;

    /* If the text entry functionality has a custom display interface,
     * does that interface support displaying a prompt string? */
    uint8_t text_has_prompt;

    /*------------------------------*/

    /* Is a touch interface available? */
    uint8_t has_touch;
};

/*------------------------- Basic functionality -------------------------*/

/**
 * sys_input_init:  Initialize input device handling functionality.  The
 * event callback function passed in as event_callback will receive input
 * events in the format defined in input.h.
 *
 * Note that event_callback may be called on an independent thread, so the
 * callback must properly lock any data it shares with the rest of the
 * library.  On some platforms, the function will be called on the same
 * thread that handles global UI processing, so the function should not
 * perform any time-consuming operations.
 *
 * Note for implementors: Touch events passed to the callback do not need
 * to have the initial_x and initial_y fields set; the callback will fill
 * them in.
 *
 * [Parameters]
 *     event_callback: Pointer to a callback function to receive input events.
 * [Return value]
 *     True on success, false on error.
 */
extern int sys_input_init(void (*event_callback)(const struct InputEvent *));

/**
 * sys_input_cleanup:  Shut down input device handling functionality.
 */
extern void sys_input_cleanup(void);

/**
 * sys_input_update:  Perform any periodic processing required to update
 * the current state of all input devices.  This routine is permitted (but
 * not required) to store input device state for returning from the other
 * sys_input_*() functions; the only requirement is that the return data
 * be no older than the last call to sys_input_update().
 */
extern void sys_input_update(void);

/**
 * sys_input_info:  Return information about available input devices.
 * This information may change if input devices are added or removed while
 * the program is running.
 *
 * [Parameters]
 *     info_ret: Pointer to structure to receive input device information.
 */
extern void sys_input_info(SysInputInfo *info_ret);

/**
 * sys_input_grab:  Activate or deactivate input grabbing.  Implements the
 * high-level function input_grab().
 *
 * [Parameters]
 *     grab: True to grab input, false to ungrab input.
 */
extern void sys_input_grab(int grab);

/**
 * sys_input_is_quit_requested:  Return whether a system-specific quit
 * request has been received.  Implements the high-level function
 * input_is_quit_requested().
 *
 * [Return value]
 *     True if a quit request has been received, false if not.
 */
extern int sys_input_is_quit_requested(void);

/**
 * sys_input_is_suspend_requested:  Return whether a system-specific
 * suspend request has been received.  Implements the high-level function
 * input_is_suspend_requested().
 *
 * [Return value]
 *     True if a suspend request has been received, false if not.
 */
extern int sys_input_is_suspend_requested(void);

/**
 * sys_input_acknowledge_suspend_request:  Acknowledge a suspend request
 * from the system, blocking until the system resumes from suspend.
 * Implements the high-level function input_acknowledge_suspend_request().
 */
extern void sys_input_acknowledge_suspend_request(void);

/*-------------------------- Joystick handling --------------------------*/

/**
 * sys_input_joystick_button_mapping:  Return the joystick button, if any,
 * which corresponds to the given button name.  Implements the high-level
 * function input_joystick_button_mapping().
 *
 * [Parameters]
 *     index: Joystick index (0 <= index < SysInputInfo.num_joysticks).
 *     name: Button name (INPUT_JOYBUTTON_*, guaranteed to be valid).
 * [Return value]
 *     Corresponding button number, or -1 if there is no corresponding button.
 */
extern int sys_input_joystick_button_mapping(int index, int name);

/**
 * sys_input_joystick_rumble:  Send a "rumble" (force feedback) effect to
 * the given joystick.  Implements the high-level function
 * input_joystick_rumble(); the parameters are guaranteed to be valid.
 *
 * [Parameters]
 *     index: Joystick index (0 <= index < SysInputInfo.num_joysticks).
 *     left: Left motor intensity (0 = off, 1 = full intensity).
 *     right: Right motor intensity (0 = off, 1 = full intensity).
 *     time: Duration of effect, in seconds (non-negative).
 */
extern void sys_input_joystick_rumble(int index, float left, float right,
                                      float time);

/*--------------------------- Mouse handling ----------------------------*/

/**
 * sys_input_mouse_set_position:  Set the position of the mouse pointer, if
 * possible.  Implements the high-level function input_mouse_set_position().
 *
 * [Parameters]
 *     x, y: New mouse pointer position.
 */
extern void sys_input_mouse_set_position(float x, float y);

/*------------------------- Text entry handling -------------------------*/

/**
 * sys_input_text_set_state:  Set the state of the text entry interface
 * (active or hidden).
 *
 * [Parameters]
 *     on: True to show the interface, false to hide it.
 *     text: Default text (NULL for no default).
 *     prompt: Prompt string to use when showing the interface (NULL for none).
 */
extern void sys_input_text_set_state(int on, const char *text,
                                     const char *prompt);

/*************************************************************************/
/*************************** Memory management ***************************/
/*************************************************************************/

/* These functions are only declared globally if a custom implementation is
 * in use (signalled by SIL_MEMORY_CUSTOM being defined).  Otherwise, the
 * declarations are limited to memory.c, but the declarations are kept here
 * to ensure consistency between the default implementation and custom
 * implementations. */

#if defined(SIL_MEMORY_CUSTOM) || defined(SIL__IN_MEMORY_C)

# ifndef SIL_MEMORY_CUSTOM
#  define extern  static inline  // Defined within memory.c.
# endif

/**
 * sys_mem_alloc:  Allocate and return a block of memory.
 *
 * The caller guarantees that size is nonzero, align is either zero or a
 * power of two, align is no greater than sys_max_mem_align(), and flags
 * does not contain MEM_ALLOC_CLEAR.
 *
 * [Parameters]
 *     size: Size of block to allocate, in bytes.
 *     align: Desired alignment of block, in bytes (always a power of 2),
 *         or zero for the system default.
 *     flags: Memory allocation flags (MEM_ALLOC_*).
 * [Return value]
 *     Pointer to allocated block, or NULL on error.
 */
extern void *sys_mem_alloc(int64_t size, int align, int flags);

/**
 * sys_mem_free:  Free a block of memory previously allocated with
 * sys_mem_alloc() or sys_mem_realloc().
 *
 * The caller guarantees that ptr is non-NULL.
 *
 * [Parameters]
 *     ptr: Pointer to block to free.
 */
extern void sys_mem_free(void *ptr);

/**
 * sys_mem_realloc:  Resize a block of memory previously allocated with
 * sys_mem_alloc() or sys_mem_realloc().
 *
 * The caller guarantees that ptr is non-NULL, size is nonzero, align is
 * equal to the alignment used when the block was allocated, and flags
 * does not contain MEM_ALLOC_CLEAR.
 *
 * [Parameters]
 *     ptr: Pointer to block to resize.
 *     size: New size for block, in bytes.
 *     align: Desired alignment of block, in bytes (always a power of 2),
 *         or zero for the system default.
 *     flags: Memory allocation flags (MEM_ALLOC_*).
 * [Return value]
 *     Pointer to resized block, or NULL on error.
 */
extern void *sys_mem_realloc(void *ptr, int64_t size, int align, int flags);

/**
 * sys_mem_avail:  Return the amount of available memory.  Implements
 * mem_avail().
 *
 * [Return value]
 *     Total amount of memory available, in bytes.
 */
extern int64_t sys_mem_avail(int flags);

/**
 * sys_mem_contig:  Return the size of the largest single allocatable block.
 * Implements mem_contig().
 *
 * [Return value]
 *     Maximum allocatable block size, in bytes.
 */
extern int64_t sys_mem_contig(int flags);

/**
 * sys_mem_max_align:  Return the maximum alignment value that can be used
 * with sys_mem_alloc().  This value must be constant over the life of the
 * program, and must be no less than sizeof(void *).
 *
 * [Return value]
 *     Maximum supported alignment, in bytes.
 */
extern CONST_FUNCTION int sys_mem_max_align(void);

# undef extern

#endif  // SIL_MEMORY_CUSTOM || SIL__IN_MEMORY_C

/*************************************************************************/
/****************** Movie (synced video/audio) playback ******************/
/*************************************************************************/

/**
 * SysMovieHandle:  Opaque handle type for a movie object.
 */
typedef struct SysMovieHandle SysMovieHandle;

/*-----------------------------------------------------------------------*/

/**
 * sys_movie_open:  Create a movie handle for a movie resource.  The
 * file handle is owned by the movie handle on success, automatically
 * closed on error.
 *
 * If direct_render is true, the function may opt not to create a texture
 * for the movie and instead use a system-dependent method to draw the
 * movie directly to the display.  In this case, sys_movie_get_texture()
 * should return zero to indicate that direct rendering will be used.
 *
 * [Parameters]
 *     fh: File handle from which to read data (guaranteed non-NULL).
 *     offset: Offset of movie data within file, in bytes.
 *     length: Size of movie data, in bytes.
 *     direct_render: True if the movie video may be rendered directly to
 *         the display, false if a texture is required.
 * [Return value]
 *     New movie handle, or NULL on error.
 */
extern SysMovieHandle *sys_movie_open(struct SysFile *fh, int64_t offset,
                                      int length, int direct_render);

/**
 * sys_movie_close:  Close the given movie handle, releasing all resources
 * used by the movie.
 *
 * [Parameters]
 *     movie: Movie handle (guaranteed non-NULL).
 */
extern void sys_movie_close(SysMovieHandle *movie);

/**
 * sys_movie_width, sys_movie_height:  Return the width or height of a
 * movie's video stream.
 *
 * [Parameters]
 *     movie: Movie handle (guaranteed non-NULL).
 * [Return value]
 *     Video width or height, in pixels.
 */
extern int sys_movie_width(SysMovieHandle *movie);
extern int sys_movie_height(SysMovieHandle *movie);

/**
 * sys_movie_framerate:  Return the frame rate for a movie's video stream.
 * Implements movie_framerate().
 *
 * [Parameters]
 *     movie: Movie handle (guaranteed non-NULL).
 * [Return value]
 *     Video frame rate, in frames per second, or zero if unknown.
 */
extern double sys_movie_framerate(SysMovieHandle *movie);

/**
 * sys_movie_set_volume:  Set the audio playback volume for a movie.
 * Implements movie_set_volume().
 *
 * [Parameters]
 *     movie: Movie handle (guaranteed non-NULL).
 *     volume: Audio volume (0...∞, 0 = silent, 1 = as recorded).
 */
void sys_movie_set_volume(SysMovieHandle *movie, float volume);

/**
 * sys_movie_play:  Start playback of a movie.  Implements movie_play().
 *
 * [Parameters]
 *     movie: Movie handle (guaranteed non-NULL).
 * [Return value]
 *     True on success, false on error.
 */
extern int sys_movie_play(SysMovieHandle *movie);

/**
 * sys_movie_stop:  Stop playback of a movie.  Implements movie_stop().
 *
 * [Parameters]
 *     movie: Movie handle (guaranteed non-NULL).
 */
extern void sys_movie_stop(SysMovieHandle *movie);

/**
 * sys_movie_get_texture:  Return the ID of a texture containing the video
 * image.  The texture ID must be constant for a given movie, but the
 * coordinates may change after each call to sys_movie_draw_frame().
 *
 * [Parameters]
 *     movie: Movie handle (guaranteed non-NULL).
 *     left_ret: Pointer to variable to receive the U coordinate of the
 *         left edge of the image (guaranteed non-NULL).
 *     right_ret: Pointer to variable to receive the U coordinate of the
 *         right edge of the image (guaranteed non-NULL).
 *     top_ret: Pointer to variable to receive the V coordinate of the
 *         top edge of the image (guaranteed non-NULL).
 *     bottom_ret: Pointer to variable to receive the V coordinate of the
 *         bottom edge of the image (guaranteed non-NULL).
 * [Return value]
 *     Texture ID, or zero if the movie was opened for direct rendering.
 */
extern int sys_movie_get_texture(
    SysMovieHandle *movie,
    float *left_ret, float *right_ret, float *top_ret, float *bottom_ret);

/**
 * sys_movie_draw_frame:  If the given movie was opened for direct rendering,
 * draw the next frame of the movie centered on the display; otherwise, draw
 * the next frame of the movie to the movie's texture.
 *
 * If the movie was opened for direct rendering, this function may freely
 * modify the graphics rendering state.
 *
 * [Parameters]
 *     movie: Movie handle (guaranteed non-NULL).
 * [Return value]
 *     True if a frame was drawn, false if the end of the stream has been
 *     reached.
 */
extern int sys_movie_draw_frame(SysMovieHandle *movie);

/*************************************************************************/
/**************************** Sound playback *****************************/
/*************************************************************************/

/**
 * SoundDecodeCallback:  Function type for the audio data decoding callback
 * function passed to sys_sound_setdata().  Decodes the next "pcm_len"
 * samples of audio data to S16LE PCM and stores them in "pcm_buffer".
 * If at least one sample but less than "pcm_len" samples are available,
 * the remaining portion of the buffer is cleared to zero.
 *
 * The output buffer must have enough room for "pcm_len" stereo samples,
 * i.e. 4*pcm_len bytes.
 *
 * Note that this function may be called in a different thread than that
 * which originally called sys_sound_setdata().
 *
 * [Parameters]
 *     handle: Audio data handle (opaque pointer).
 *     pcm_buffer: Output buffer for PCM samples.
 *     pcm_len: Number of samples to retrieve.
 * [Return value]
 *     True on success, false on end of stream or error.
 */
typedef int (*SoundDecodeCallback)(void *handle, void *pcm_buffer, int pcm_len);

/*---------------------- General control functions ----------------------*/

/**
 * sys_sound_init:  Initialize the audio output functionality, and allocate
 * an array of playback channels.  The channel IDs (arbitrary unique
 * integers) to be used in calling other sys_sound_*() functions are stored
 * consecutively in the array pointed to by "channels_ret".
 *
 * This function may only be called while the audio output functionality
 * is not initialized, i.e. either at program startup or after a
 * sys_sound_cleanup() call.
 *
 * [Parameters]
 *     device_name: System-dependent device name for audio output.
 *     num_channels: Number of playback channels to allocate.
 *     channels_ret: Array into which channel IDs are stored.
 * [Return value]
 *     Number of channels available for playback, or zero on error.
 */
extern int sys_sound_init(const char *device_name, int num_channels,
                          int *channels_ret);

/**
 * sys_sound_playback_rate:  Return the system's sampling rate for audio
 * playback.  A return value of zero indicates that the system can accept
 * data at any sampling rate; any other value will cause all audio data to
 * be resampled to the specified rate (if necessary) during decoding.
 *
 * The value returned by this function must remain constant between a call
 * to sys_sound_init() and the corresponding call to sys_sound_cleanup().
 *
 * [Return value]
 *     Playback sampling rate, in Hz, or zero to accept any sampling rate.
 */
extern int sys_sound_playback_rate(void);

/**
 * sys_sound_check_headphone_disconnect:  Return true if the a headphone
 * disconnect event has been detected.  Implements the high-level function
 * sound_check_headphone_disconnect().
 *
 * This function will not be called (and may be left undefined) if
 * SIL_SOUND_HEADPHONE_CHECK is not defined.
 *
 * [Return value]
 *     True if a headphone disconnect has been detected, false if not.
 */
extern int sys_sound_check_headphone_disconnect(void);

/**
 * sys_sound_acknowledge_headphone_disconnect:  Reset the headphone-
 * disconnect flag.  Implements the high-level function
 * sound_acknowledge_headphone_disconnect().
 *
 * This function will not be called (and may be left undefined) if
 * SIL_SOUND_HEADPHONE_CHECK is not defined.
 */
extern void sys_sound_acknowledge_headphone_disconnect(void);

/**
 * sys_sound_cleanup:  Shut down the audio output functionality.
 */
extern void sys_sound_cleanup(void);

/*---------------- Sound playback/manipulation functions ----------------*/

enum SoundFormat;  // Defined in sound.h

/**
 * sys_sound_setdata:  Register an audio data stream on a sound channel.
 * This function fails if a stream has already been registered on the
 * channel; to reuse a channel, first call sys_sound_reset() to clear the
 * existing stream.
 *
 * Decoding may begin as soon as this function is called, so volume and
 * panning should be set before calling this function.
 *
 * If sys_sound_playback_rate() returns a nonzero value, sample_rate is
 * guaranteed to be equal to that value.  In this case, any needed
 * resampling is handled by the decode function.
 *
 * [Parameters]
 *     channel: Channel ID for playback.
 *     decode_func: Decode callback function for audio data.
 *     handle: Audio data handle passed to decode function.
 *     is_stereo: True if the audio data is stereo, false if monaural.
 *     sample_rate: Sampling rate of data returned from decode function, in Hz.
 * [Return value]
 *     True on success, false on error.
 */
extern int sys_sound_setdata(int channel, SoundDecodeCallback decode_func,
                             void *handle, int is_stereo, int sample_rate);

/**
 * sys_sound_setvol:  Set the playback volume on the given channel.  Any
 * fade effect on the channel is cancelled by this call.
 *
 * This function may be called whether data has been registered on the
 * channel or not.
 *
 * [Parameters]
 *     channel: Channel ID.
 *     volume: Playback volume (0...∞, 0 = silent, 1 = as recorded).
 */
extern void sys_sound_setvol(int channel, float volume);

/**
 * sys_sound_setpan:  Set the pan position for the given channel.  For
 * stereo samples, the channel opposite the pan direction (e.g., the right
 * channel when panning left) is shifted in the direction of the pan, and
 * the overall volume is scaled down by a factor of 1+|pan|.
 *
 * This function may be called whether data has been registered on the
 * channel or not.
 *
 * [Parameters]
 *     channel: Channel ID.
 *     pan: Pan position (-1 = left, 0 = center, +1 = right).
 */
extern void sys_sound_setpan(int channel, float pan);

/**
 * sys_sound_setfade:  Begin or cancel a volume fade on the given channel.
 * Note that attempting to cancel a fade started with cut != 0 may trigger
 * a race condition, and should thus be avoided.
 *
 * [Parameters]
 *     channel: Channel ID.
 *     target: Target volume (0...∞, 0 = silent, 1 = as recorded).
 *     length: Fade time, in seconds, or zero to cancel any current fade.
 *     cut: True to terminate playback when the volume reaches zero;
 *         false to continue playback regardless of volume.
 */
extern void sys_sound_setfade(int channel, float target, float length, int cut);

/**
 * sys_sound_start:  Begin playing the given channel's audio data.  If
 * playback was previously stopped with sys_sound_stop(), playback resumes
 * from the point at which it was stopped.  This function does nothing if
 * no audio data has been registered on the channel.
 *
 * [Parameters]
 *     channel: Channel ID.
 */
extern void sys_sound_start(int channel);

/**
 * sys_sound_stop:  Stop playback on the given channel.  This function
 * does nothing if the channel is not currently playing.
 *
 * [Parameters]
 *     channel: Channel ID.
 */
extern void sys_sound_stop(int channel);

/**
 * sys_sound_reset:  Stop playback on the given channel and clear any
 * registered audio data, waiting until playback has actually stopped
 * before returning.
 *
 * [Parameters]
 *     channel: Channel ID.
 */
extern void sys_sound_reset(int channel);

/**
 * sys_sound_status:  Return whether the given channel is currently playing.
 *
 * [Parameters]
 *     channel: Channel ID.
 * [Return value]
 *     True if the channel is playing, false if not.
 */
extern int sys_sound_status(int channel);

/*************************************************************************/
/*************************** Texture handling ****************************/
/*************************************************************************/

/**
 * SysTexture:  Opaque type for system-specific texture data.
 */
typedef struct SysTexture SysTexture;

/**
 * SysTextureLockMode (SYS_TEXTURE_LOCK_*):  Values for the lock_mode
 * parameter to sys_texture_lock().
 */
enum SysTextureLockMode {
    /* Return an array of 32-bit RGBA pixels. */
    SYS_TEXTURE_LOCK_NORMAL = 1,

    /* Return an uninitialized buffer for 32-bit RGBA pixels. */
    SYS_TEXTURE_LOCK_DISCARD,
};
typedef enum SysTextureLockMode SysTextureLockMode;

/* Invalid SysTextureLockMode value that can be used as a placeholder.
 * Defined outside the enum so we don't get warnings about it for -Wswitch. */
#define SYS_TEXTURE_LOCK_INVALID ((SysTextureLockMode)0)

/*-------------------- Texture creation and deletion --------------------*/

/**
 * sys_texture_create:  Create a new SysTexture object, and fill it with
 * the provided image data, if any.
 *
 * If mipmaps != 0, the image data provided is used only for the first
 * (full-size) level, and any subsequent mipmap levels in the data are
 * ignored.
 *
 * If reuse != 0, then the memory region pointed to by data (if data != NULL)
 * must have been allocated with mem_alloc().  On success, the data buffer
 * will be managed by the SysTexture object (or freed if it can't feasibly
 * be reused); on error, this function will free the data buffer.
 *
 * If reuse == 0, the memory region pointed to by data will not be modified.
 *
 * [Parameters]
 *     width: Texture width, in pixels (guaranteed >=1).
 *     height: Texture height, in pixels (guaranteed >=1).
 *     data_format: Image data format (one of TEX_FORMAT_* from
 *         utility/tex-file.h).
 *     num_levels: Number of mipmap levels (including the base image).
 *         Zero indicates that the texture should be created empty.
 *     data: Pointer to image data.  Ignored if num_levels == 0.
 *     stride: Line size of image data for base level, in pixels.  Ignored
 *         if num_levels == 0 or if the image data format is a compressed
 *         format.
 *     level_offsets: Byte offsets into data for each image level;
 *         level_offsets[0] points to the main image data and includes the
 *         palette data for paletted textures.  Ignored if num_levels == 0.
 *     level_sizes: Size of data, in bytes, for each image level.  Ignored
 *         if num_levels == 0.
 *     mipmaps: True to enable automatic mipmap generation for this texture
 *         (if applicable to the system); false to prevent mipmap generation.
 *     mem_flags: Memory allocation flags (MEM_ALLOC_*).
 *     reuse: True to reuse the image data, or free it if reuse is not
 *         possible; false to copy the image data.
 * [Return value]
 *     New texture, or NULL on error.
 */
extern SysTexture *sys_texture_create(
    int width, int height, enum TextureFormat data_format, int num_levels,
    void *data, int stride, const int32_t *level_offsets,
    const int32_t *level_sizes, int mipmaps, int mem_flags, int reuse);

/**
 * sys_texture_destroy:  Destroy a SysTexture object.
 *
 * [Parameters]
 *     texture: Texture to destroy (guaranteed non-NULL and not locked).
 */
extern void sys_texture_destroy(SysTexture *texture);

/*-------------------- Texture information retrieval --------------------*/

/**
 * sys_texture_width, sys_texture_height:  Return the width or height of
 * the given texture.
 *
 * [Parameters]
 *     texture: Texture for which to retrieve size.
 * [Return value]
 *     Width or height of the texture, in pixels.
 */
extern int sys_texture_width(SysTexture *texture);
extern int sys_texture_height(SysTexture *texture);

/**
 * texture_has_mipmaps:  Return whether the given texture has stored
 * mipmaps in addition to the base image.
 *
 * [Parameters]
 *     texture: Texture for which to retrieve mipmap state.
 * [Return value]
 *     True if the texture has stored mipmaps, false if not.
 */
extern int sys_texture_has_mipmaps(SysTexture *texture);

/*----------------------- Pixel data manipulation -----------------------*/

/**
 * sys_texture_grab:  Overwrite all or part of the given texture's image
 * data with a portion of the display.  Implements the high-level function
 * texture_grab().
 *
 * [Parameters]
 *     texture: Texture into which to copy display data.
 *     x, y: Base display coordinates of region to copy, in pixels.
 *     w, h: Size of region to copy, in pixels.
 *     readable: False if the texture is not required to be readable.
 * [Return value]
 *     True on success, false on error.
 */
extern int sys_texture_grab(SysTexture *texture, int x, int y, int w, int h,
                            int readable);

/**
 * sys_texture_lock:  Lock a texture's image data into memory, and return
 * a pointer to all or some of the texture's pixel data.  Implements the
 * high-level texture_lock() family of functions.  The caller guarantees
 * that the texture is not already locked.
 *
 * For SYS_TEXTURE_LOCK_RAW, the position and size coordinates are ignored.
 *
 * [Parameters]
 *     texture: Texture to lock.
 *     lock_mode: Lock mode (SYS_TEXTURE_LOCK_*).
 *     x, y: Base texture coordinates of region to lock, in pixels.
 *     w, h: Size of region to lock, in pixels.
 * [Return value]
 *     Pointer to pixel data, or NULL on error.
 */
extern uint8_t *sys_texture_lock(SysTexture *texture,
                                 SysTextureLockMode lock_mode,
                                 int x, int y, int w, int h);

/**
 * sys_texture_unlock:  Unlock a texture locked with sys_texture_lock().
 * If update != 0, the texture is updated from the data in the pixel data
 * buffer returned by sys_texture_lock().  The caller guarantees that the
 * texture is currently locked.
 *
 * [Parameters]
 *     texture: Texture to unlock.
 *     update: True to copy the pixel data into the texture; false to
 *         leave the texture data as is.
 */
extern void sys_texture_unlock(SysTexture *texture, int update);

/**
 * sys_texture_flush:  Assume that texture data stored in memory may have
 * been modified, and ensure such modifications are visible to the graphics
 * hardware.  This function is called by texture_unlock() when unlocking
 * after a texture_lock_raw() call, and will never be called while the
 * texture is locked with sys_texture_lock().
 *
 * [Parameters]
 *     texture: Texture to flush.
 */
extern void sys_texture_flush(SysTexture *texture);

/*------------------------ Rendering operations -------------------------*/

/**
 * sys_texture_set_repeat:  Set the given texture's coordinate wrap flags.
 *
 * [Parameters]
 *     texture: Texture to modify.
 *     repeat_u: True to repeat horizontally, false to clamp.
 *     repeat_v: True to repeat vertically, false to clamp.
 */
extern void sys_texture_set_repeat(SysTexture *texture, int repeat_u, int repeat_v);

/**
 * sys_texture_set_antialias:  Set the given texture's antialiasing flag.
 *
 * [Parameters]
 *     texture: Texture to modify.
 *     on: True to enable antialiasing, false to disable.
 */
extern void sys_texture_set_antialias(SysTexture *texture, int on);

/**
 * sys_texture_apply:  Set the given texture as the texture to be used for
 * subsequent rendering operations.
 *
 * [Parameters]
 *     texture: Texture to set, or NULL to clear any previously set texture.
 */
extern void sys_texture_apply(SysTexture *texture);

/*************************************************************************/
/***************** Thread management and synchronization *****************/
/*************************************************************************/

/*----------------------------- Data types ------------------------------*/

/**
 * SysThreadID:  Thread handle type.  Note that this and other low-level
 * synchronization primitive types are pointer-sized integers rather than
 * actual pointers to accommodate systems which use a numeric system ID
 * rather than allocating memory for the object.
 */
typedef uintptr_t SysThreadID;

/**
 * SysMutexID:  Mutex (mutual-exclusion primitive) type.
 */
typedef uintptr_t SysMutexID;

/**
 * SysSemaphoreID:  Semaphore type.
 */
typedef uintptr_t SysSemaphoreID;

/*--------------------------- Thread routines ---------------------------*/

/**
 * sys_thread_create:  Create and start a thread executing the given
 * function at the given priority and with the given stack size.  Implements
 * the high-level functions thread_create(), thread_create_with_priority(),
 * and thread_create_with_stack_size().
 *
 * [Parameters]
 *     stack_size: Stack size, in bytes, or 0 for the default stack size
 *         (guaranteed nonnegative).
 *     priority: Thread priority.
 *     function: Function to execute (guaranteed non-NULL).
 *     param: Arbitrary parameter to pass to the function.
 * [Return value]
 *     New thread handle, or zero on error.
 */
extern SysThreadID sys_thread_create(
    int priority, int stack_size, int (*function)(void *), void *param);

/**
 * sys_thread_is_running:  Return whether the given thread is still
 * running.  Implements the high-level function thread_is_running().
 *
 * [Parameters]
 *     thread: Handle of thread to check (guaranteed nonzero).
 * [Return value]
 *     True if the thread is still running, false if the thread has terminated.
 */
extern int sys_thread_is_running(SysThreadID thread);

/**
 * sys_thread_wait:  Wait for the given thread to terminate, and return
 * its result value.  Implements the high-level functions thread_wait() and
 * thread_wait2().
 *
 * [Parameters]
 *     thread: Handle of thread to wait for (guaranteed nonzero).
 *     result_ret: Pointer to variable to receive thread result value
 *         (guaranteed non-NULL).
 * [Return value]
 *     True on success, false on error.
 */
extern int sys_thread_wait(SysThreadID thread, int *result_ret);

/**
 * sys_thread_yield:  Yield the CPU to another thread.  Implements the
 * high-level function thread_yield().
 */
extern void sys_thread_yield(void);

/*--------------------------- Mutex routines ----------------------------*/

/**
 * sys_mutex_create:  Create a mutex.  Implements the high-level function
 * mutex_create().
 *
 * [Parameters]
 *     initially_locked: True to create the mutex in a locked state, false
 *         to create it in an unlocked state.
 * [Return value]
 *     New mutex handle, or zero on error.
 */
extern SysMutexID sys_mutex_create(int initially_locked);

/**
 * sys_mutex_destroy:  Destroy a mutex.  Implements the high-level function
 * mutex_destroy().
 *
 * [Parameters]
 *     mutex: Mutex to destroy (guaranteed nonzero).
 */
extern void sys_mutex_destroy(SysMutexID mutex);

/**
 * sys_mutex_lock:  Lock a mutex.  Implements the high-level function
 * mutex_lock().
 *
 * [Parameters]
 *     mutex: Mutex to lock (guaranteed nonzero).
 */
extern void sys_mutex_lock(SysMutexID mutex);

/**
 * sys_mutex_lock_timeout:  Lock a mutex if it becomes available within the
 * specified timeout.  Implements the high-level function mutex_lock_timeout().
 *
 * [Parameters]
 *     mutex: Mutex to lock (guaranteed nonzero).
 *     timeout: Maximum time to wait for the mutex, in seconds (guaranteed
 *         nonnegative).
 * [Return value]
 *     True if the lock operation succeeded, false if not.
 */
extern int sys_mutex_lock_timeout(SysMutexID mutex, float timeout);

/**
 * sys_mutex_unlock:  Unlock a mutex.  Implements the high-level function
 * mutex_unlock().
 *
 * [Parameters]
 *     mutex: Mutex to unlock (guaranteed nonzero).
 */
extern void sys_mutex_unlock(SysMutexID mutex);

/*------------------------- Semaphore routines --------------------------*/

/**
 * sys_semaphore_create:  Create a semaphore.  Implements the high-level
 * function semaphore_create().
 *
 * [Parameters]
 *     initial_value: Initial value for the semaphore (guaranteed nonnegative).
 *     required_max: Required maximum value for the semaphore (guaranteed
 *         positive and no less than initial_value).
 * [Return value]
 *     New signal handle, or zero on error.
 */
extern SysSemaphoreID sys_semaphore_create(int initial_value, int required_max);

/**
 * sys_semaphore_destroy:  Destroy a semaphore.  Implements the high-level
 * function semaphore_destroy().
 *
 * [Parameters]
 *     semaphore: Semaphore to destroy (guaranteed nonzero).
 */
extern void sys_semaphore_destroy(SysSemaphoreID semaphore);

/**
 * sys_semaphore_wait:  Wait on a semaphore.  Implements the high-level
 * function semaphore_wait().
 *
 * [Parameters]
 *     semaphore: Semaphore on which to wait (guaranteed nonzero).
 */
extern void sys_semaphore_wait(SysSemaphoreID semaphore);

/**
 * sys_semaphore_wait_timeout:  Wait on a semaphore for up to the specified
 * timeout.  Implements the high-level function semaphore_wait_timeout().
 *
 * [Parameters]
 *     semaphore: Semaphore on which to wait (guaranteed nonzero).
 *     timeout: Maximum time to wait for the signal, in seconds
 *         (guaranteed nonnegative).
 * [Return value]
 *     True if the lock operation succeeded, false if not.
 */
extern int sys_semaphore_wait_timeout(SysSemaphoreID semaphore, float timeout);

/**
 * sys_semaphore_signal:  Signal a semaphore.  Implements the high-level
 * function semaphore_signal().
 *
 * [Parameters]
 *     semaphore: Semaphore to signal (guaranteed nonzero).
 */
extern void sys_semaphore_signal(SysSemaphoreID semaphore);

/*************************************************************************/
/*********************** Timekeeping functionality ***********************/
/*************************************************************************/

/**
 * sys_time_init:  Initialize timekeeping functionality, and return the
 * number of time units per second in the time values returned by
 * sys_time_now().  This function must not fail.
 *
 * This function may be called more than once; the implementation is free
 * to reset internal counters or other data on each call (in particular,
 * the time unit and values returned from sys_time_now() may be reset by a
 * call to this function).
 *
 * [Return value]
 *     Number of time units per second.
 */
extern int sys_time_init(void);

/**
 * sys_time_now:  Return the system time.  The meaning of the value is
 * system-dependent; the value should only be used for comparison or
 * difference calculations with other time values returned from this
 * function.
 *
 * Implementations may assume that the value returned from the first call
 * to sys_time_now() following a call to sys_time_init() will be used as
 * the epoch for values reported from time_now().  For example, the POSIX
 * implementation of these functions exports this value via the
 * sys_posix_time_epoch() function, allowing other system timestamps to be
 * converted directly to floating-point values compatible with time_now().
 *
 * [Return value]
 *     Current time, in units as returned by the last call to sys_time_init().
 */
extern int64_t sys_time_now(void);

/**
 * sys_time_delay:  Delay the program for the given length of time (which
 * must be less than or equal to one minute).  Depending on the system, the
 * program may actually delay for slightly longer, but will never return
 * before at least the specified amount of time has passed.
 *
 * The behavior of this routine is undefined if the time value given
 * corresponds to a length of time greater than 60 seconds.
 *
 * [Parameters]
 *     time: Length of time to wait, in units as returned by the last call
 *         to sys_time_init().
 */
extern void sys_time_delay(int64_t time);

/*************************************************************************/
/*************************** User data access ****************************/
/*************************************************************************/

/* Structures defined in userdata.h to which pointers are used here. */
struct UserStatInfo;
struct UserStatValue;

/**
 * SysUserDataOp:  Enumeration of user data operations for the
 * SysUserDataParams.operation field.
 */
enum SysUserDataOp {
    SYS_USERDATA_SAVE_SAVEFILE = 1,
    SYS_USERDATA_LOAD_SAVEFILE,
    SYS_USERDATA_DELETE_SAVEFILE,
    SYS_USERDATA_SCAN_SAVEFILES,
    SYS_USERDATA_SAVE_SETTINGS,
    SYS_USERDATA_LOAD_SETTINGS,
    SYS_USERDATA_SAVE_SCREENSHOT,
    SYS_USERDATA_SAVE_DATA,
    SYS_USERDATA_LOAD_DATA,
    SYS_USERDATA_DELETE_DATA,
    SYS_USERDATA_REGISTER_STATS,
    SYS_USERDATA_UPDATE_STATS,
    SYS_USERDATA_RETRIEVE_STATS,
    SYS_USERDATA_CLEAR_STATS,
};
typedef enum SysUserDataOp SysUserDataOp;

/**
 * SysUserDataParamsPrivate:  Opaque structure which may be used by system
 * implementations to store private data associated with a SysUserDataParams
 * parameter block.
 */
typedef struct SysUserDataParamsPrivate SysUserDataParamsPrivate;

/**
 * SysUserDataParams:  Structure holding parameters for a user data
 * operation.
 */
typedef struct SysUserDataParams SysUserDataParams;
struct SysUserDataParams {
    /* The operation to be performed (one of SYS_USERDATA_*). */
    SysUserDataOp operation;

    /* Override pathname for this operation (overrides the default
     * system-dependent pathname), or NULL if none. */
    const char *override_path;

    /* Program name or ID string (required for all operations). */
    const char *program_name;
    /* Game title (required for all operations). */
    const char *game_title;

    /* Save file number (SAVE_SAVEFILE, LOAD_SAVEFILE, SCAN_SAVEFILES,
     * DELETE_SAVEFILE only).  For SCAN_SAVEFILES, this is the number of
     * the first save to scan for. */
    int savefile_num;
    /* Data file path (SAVE_DATA, LOAD_DATA, DELETE_DATA only). */
    const char *datafile_path;
    /* File title and descriptive text (SAVE_SAVEFILE, SAVE_SETTINGS only). */
    const char *title;
    const char *desc;

    /* Pointer to the data to be saved (SAVE operations except
     * SAVE_SCREENSHOT only).  Must remain valid until the operation
     * completes. */
    const void *save_data;
    /* Size in bytes of the data to be saved (SAVE operations except
     * SAVE_SCREENSHOT only). */
    int save_size;
    /* Pointer to the RGBA pixel data to be saved as a screenshot
     * (SAVE_SAVEFILE, SAVE_SCREENSHOT only).  Must remain valid until the
     * operation completes. */
    const void *save_image;
    /* Size in pixels of the screenshot data to be saved (SAVE_SAVEFILE,
     * SAVE_SCREENSHOT only). */
    int save_image_width, save_image_height;

    /* Output field, set to a pointer to a buffer allocated with mem_alloc()
     * containing the loaded data (LOAD operations only, undefined on
     * operation failure). */
    void *load_data;
    /* Output field, set to the size in bytes of the loaded data (LOAD
     * operations only, undefined on operation failure). */
    int load_size;
    /* Output field, set to a pointer to a buffer allocated with mem_alloc()
     * containing the loaded RGBA pixel data for the save file's screenshot,
     * or NULL if no screenshot was found (LOAD_SAVEFILE only, undefined on
     * operation failure). */
    void *load_image;
    /* Output field, set to the size in pixels of the loaded screenshot
     * data (LOAD_SAVEFILE only, undefined on operation failure). */
    int load_image_width, load_image_height;

    /* Pointer to a buffer to be filled with savefile scan results
     * (SCAN_SAVEFILES only). */
    uint8_t *scan_buffer;
    /* Number of files to scan for (SCAN_SAVEFILES only).  scan_buffer must
     * have room for at least this many entries. */
    int scan_count;

    /* Pointer to and length of an array of UserStatInfo structures to
     * register (REGISTER_STATS only).  This must include every statistic
     * to be recorded.  The array need not remain valid after
     * sys_userdata_start() returns. */
    const struct UserStatInfo *stat_info;
    int stat_info_len;

    /* Pointer to and length of an array of UserStatValue structures to
     * update (UPDATE_STATS only).  This should include only the
     * statistics to be updated.  The array need not remain valid after
     * sys_userdata_start() returns. */
    const struct UserStatValue *stat_values;
    int stat_values_len;

    /* Pointer to and length of an array of UserStatValue structures to
     * store statistic values in (RETRIEVE_STATS only).  This should
     * include only the statistics to be retrieved.  The array must remain
     * valid until the operation completes; if the operation fails, the
     * values stored in the arrays will be undefined. */
    struct UserStatValue *stat_buffer;
    int stat_buffer_len;

    /* Pointer to implementation-private data (opaque to callers). */
    SysUserDataParamsPrivate *private;
};

/*-----------------------------------------------------------------------*/

/**
 * sys_userdata_get_data_path:  Return a pathname suitable for accessing
 * resources stored as user data, as for the system-independent interface
 * function userdata_get_data_path().  The returned string (if any) is
 * stored in a buffer allocated with mem_alloc(), which should be freed
 * with mem_free() when no longer needed.
 *
 * [Parameters]
 *     program_name: Program name for which to generate data pathname.
 * [Return value]
 *     Newly-allocated buffer containing generated pathname; NULL on error
 *     or if not supported by this platform.
 */
extern char *sys_userdata_get_data_path(const char *program_name);

/**
 * sys_userdata_perform:  Perform the user data operation specified by the
 * passed-in parameter block.
 *
 * The platform-independent user data interface is asynchronous to allow
 * the caller to perform other actions (such as updating the display)
 * while a user data operation is in progress.  This is implemented by
 * performing each user data operation in its own thread, so this function
 * must take care to protect any shared data with mutexes or similar
 * locking mechanisms.
 *
 * [Parameters]
 *     params: Parameter block describing operation to be performed.
 * [Return value]
 *     True if the operation was successfully performed, false if an error
 *     occurred.
 */
extern int sys_userdata_perform(SysUserDataParams *params);

/*************************************************************************/
/********************** Miscellaneous functionality **********************/
/*************************************************************************/

/**
 * sys_display_error:  Display an error message to the user.  Implements
 * the high-level utility function display_error().
 *
 * [Parameters]
 *     message: Error message to display (guaranteed non-NULL).  May
 *         contain printf()-style formatting tokens.
 *     args: Format arguments for error message.
 */
extern void sys_display_error(const char *message, va_list args);

/**
 * sys_get_language:  Return the system's preferred language as a 2-letter
 * ISO 639-1 language code (e.g., "en" for English), along with an optional
 * 2-letter ISO 3166 region code identifying the dialect (e.g., "UK" for
 * British English).  Implements the high-level utility function
 * get_system_language().
 *
 * [Parameters]
 *     index: Language preference index (0 = system default language).
 *     language_ret: Pointer to string buffer to receive the language code.
 *     dialect_ret: Pointer to string buffer to receive the dialect code.
 * [Return value]
 *     True on success, false if there is no language for the requested index.
 */
extern int sys_get_language(int index, char *language_ret, char *dialect_ret);

/**
 * sys_last_error:  Return a numeric code describing the error associated
 * with the previous failing system-specific function call.  Behavior is
 * undefined if not called immediately after such a failed call.
 *
 * [Return value]
 *     Error code (SYSERR_*).
 */
extern int sys_last_error(void);
enum {
    SYSERR_UNKNOWN_ERROR = 1,   // All errors except the below.
    SYSERR_INVALID_PARAMETER,   // An invalid parameter value was used.
    SYSERR_OUT_OF_MEMORY,       // Insufficient memory (or other system
                                //    resource) is available.
    SYSERR_BUFFER_OVERFLOW,     // An internal buffer size was exceeded.
    SYSERR_TRANSIENT_FAILURE,   // A transient failure occurred; retrying the
                                //    operation at a later time may succeed.
    SYSERR_FILE_NOT_FOUND,      // The file or directory could not be opened.
    SYSERR_FILE_ACCESS_DENIED,  // Access to the file or directory was denied.
    SYSERR_FILE_WRONG_TYPE,     // The file is of the wrong type (for example,
                                //    sys_file_open() on a directory).
    SYSERR_FILE_ASYNC_ABORTED,  // The asynchronous read was aborted.
    SYSERR_FILE_ASYNC_INVALID,  // The asynchronous read ID is invalid.
    SYSERR_FILE_ASYNC_FULL,     // Too many asynchronous reads are in progress.
};

/**
 * sys_last_errstr:  Return a text string describing the error associated
 * with the previous failing system-specific function call.  Behavior is
 * undefined if not called immediately after such a failed call, except
 * that the return value will always be a valid string.
 *
 * [Return value]
 *     Error description.
 */
extern const char *sys_last_errstr(void);

/**
 * sys_open_file:  Open a local file in an appropriate external program;
 * if path == NULL, just return whether opening of local files is supported
 * by the system, under the assumption that the file type is recognized.
 * Implements the high-level utility functions can_open_file() and open_file().
 *
 * [Parameters]
 *     path: Path of file to open, or NULL to check for file opening
 *         capability.
 * [Return value]
 *     False if the system does not support opening files in external
 *     applications, or if path != NULL and an error is known to have
 *     occurred; true otherwise.
 */
extern int sys_open_file(const char *path);

/**
 * sys_open_url:  Open a URL in the user's browser; if url == NULL, just
 * return whether web browsing is supported by the system.  Implements the
 * high-level utility functions can_open_url() and open_url().
 *
 * [Parameters]
 *     url: URL to open, or NULL to check for web browsing capability.
 * [Return value]
 *     False if the system does not support web browsing, or if url != NULL
 *     and an error is known to have occurred; true otherwise.
 */
extern int sys_open_url(const char *url);

/**
 * sys_random_seed:  Return a nonzero random seed derived from the runtime
 * environment or other state.  The returned value should be different
 * across multiple runs of the program.
 *
 * This function must _not_ call any of the random number generation
 * functions.
 *
 * [Return value]
 *     Nonzero random seed.
 */
extern uint64_t sys_random_seed(void);

/**
 * sys_reset_idle_timer:  Reset any system "idle timers", such as
 * screensaver activation or auto-suspend timers.  Implements the
 * high-level utility function reset_idle_timer().
 */
extern void sys_reset_idle_timer(void);

/**
 * sys_set_performance_level:  Request the system to switch to a
 * different performance level.  Implements the high-level function
 * set_performance_level().
 *
 * [Parameters]
 *     level: New performance level (PERFORMANCE_LEVEL_* or a
 *         system-dependent positive value).
 * [Return value]
 *     True on success, false on error.
 */
extern int sys_set_performance_level(int level);

#if defined(DEBUG) && defined(SIL_DLOG_CUSTOM)
/**
 * sys_write_DLOG:  Output a debugging message generated by the DLOG()
 * macro.
 *
 * The function may destroy the contents of "message".  In particular, the
 * caller guarantees that modifying message[len] (for example, to store a
 * newline character) is safe.
 *
 * Only defined in debugging mode.
 *
 * [Parameters]
 *     message: Message to output.
 *     len: Length of message, in bytes.
 */
extern void sys_write_DLOG(char *message, int len);
#endif

/*************************************************************************/
/************************ Debug interface helpers ************************/
/*************************************************************************/

/**
 * debug_get_memory_stats:  Retrieve system memory usage statistics.
 * Values returned are in units of bytes.
 *
 * [Parameters]
 *     total_ret: Pointer to variable to receive the total amount of memory
 *         in the system.
 *     self_ret: Pointer to variable to receive the amount of memory used
 *         by the current process.
 *     avail_ret: Pointer to variable to receive amount of memory available
 *         in the system.
 * [Return value]
 *     True on success, false on error.
 */
extern int sys_debug_get_memory_stats(
    int64_t *total_ret, int64_t *self_ret, int64_t *avail_ret);

/*************************************************************************/
/********************** Macros for testing (part 2) **********************/
/*************************************************************************/

#if defined(INCLUDE_TESTS) && !defined(IN_SYSDEP) && !defined(IN_SYSDEP_TEST)
// Through the end of this section.

extern int is_running_tests(void);

#define DEFINE_STUB(type,name,params,args) \
extern type TEST_##name params; \
static inline type STUB_##name params { \
    return is_running_tests() ? TEST_##name args : name args; \
}

DEFINE_STUB(int, sys_input_init, (void (*event_callback)(const struct InputEvent *)), (event_callback))
DEFINE_STUB(void, sys_input_cleanup, (void), ())
DEFINE_STUB(void, sys_input_update, (void), ())
DEFINE_STUB(void, sys_input_info, (SysInputInfo *info_ret), (info_ret))
DEFINE_STUB(void, sys_input_grab, (int grab), (grab))
DEFINE_STUB(int, sys_input_is_quit_requested, (void), ())
DEFINE_STUB(int, sys_input_is_suspend_requested, (void), ())
DEFINE_STUB(void, sys_input_acknowledge_suspend_request, (void), ())
DEFINE_STUB(int, sys_input_joystick_button_mapping, (int index, int name), (index, name))
DEFINE_STUB(void, sys_input_joystick_rumble, (int index, float left, float right, float time), (index, left, right, time))
DEFINE_STUB(void, sys_input_mouse_set_position, (float x, float y), (x, y))
DEFINE_STUB(void, sys_input_text_set_state, (int on, const char *text, const char *prompt), (on, text, prompt))

DEFINE_STUB(int, sys_sound_init, (const char *device_name, int num_channels, int *channels_ret), (device_name, num_channels, channels_ret))
DEFINE_STUB(int, sys_sound_playback_rate, (void), ())
DEFINE_STUB(int, sys_sound_check_headphone_disconnect, (void), ())
DEFINE_STUB(void, sys_sound_acknowledge_headphone_disconnect, (void), ())
DEFINE_STUB(void, sys_sound_cleanup, (void), ())
DEFINE_STUB(int, sys_sound_setdata, (int channel, SoundDecodeCallback decode_func, void *handle, int is_stereo, int sample_rate), (channel, decode_func, handle, is_stereo, sample_rate))
DEFINE_STUB(void, sys_sound_setvol, (int channel, float volume), (channel, volume))
DEFINE_STUB(void, sys_sound_setpan, (int channel, float pan), (channel, pan))
DEFINE_STUB(void, sys_sound_setfade, (int channel, float target, float length, int cut), (channel, target, length, cut))
DEFINE_STUB(void, sys_sound_start, (int channel), (channel))
DEFINE_STUB(void, sys_sound_stop, (int channel), (channel))
DEFINE_STUB(void, sys_sound_reset, (int channel), (channel))
DEFINE_STUB(int, sys_sound_status, (int channel), (channel))

DEFINE_STUB(int, sys_time_init, (void), ())
DEFINE_STUB(int64_t, sys_time_now, (void), ())
DEFINE_STUB(void, sys_time_delay, (int64_t time), (time))

DEFINE_STUB(char *, sys_userdata_get_data_path, (const char *program_name), (program_name))
DEFINE_STUB(int, sys_userdata_perform, (SysUserDataParams *params), (params))

DEFINE_STUB(void, sys_display_error, (const char *message, va_list args), (message, args))
DEFINE_STUB(int, sys_get_language, (int index, char *language_ret, char *dialect_ret), (index, language_ret, dialect_ret))
DEFINE_STUB(int, sys_open_file, (const char *path), (path))
DEFINE_STUB(int, sys_open_url, (const char *url), (url))
DEFINE_STUB(void, sys_reset_idle_timer, (void), ())

DEFINE_STUB(int, sys_debug_get_memory_stats, (int64_t *total_ret, int64_t *self_ret, int64_t *avail_ret), (total_ret, self_ret, avail_ret))

#undef DEFINE_STUB

/*-----------------------------------------------------------------------*/

#define sys_input_init                  STUB_sys_input_init
#define sys_input_cleanup               STUB_sys_input_cleanup
#define sys_input_update                STUB_sys_input_update
#define sys_input_info                  STUB_sys_input_info
#define sys_input_grab                  STUB_sys_input_grab
#define sys_input_is_quit_requested     STUB_sys_input_is_quit_requested
#define sys_input_is_suspend_requested  STUB_sys_input_is_suspend_requested
#define sys_input_acknowledge_suspend_request STUB_sys_input_acknowledge_suspend_request
#define sys_input_joystick_button_mapping STUB_sys_input_joystick_button_mapping
#define sys_input_joystick_rumble       STUB_sys_input_joystick_rumble
#define sys_input_mouse_set_position    STUB_sys_input_mouse_set_position
#define sys_input_text_set_state        STUB_sys_input_text_set_state

#define sys_sound_init                  STUB_sys_sound_init
#define sys_sound_playback_rate         STUB_sys_sound_playback_rate
#define sys_sound_check_headphone_disconnect STUB_sys_sound_check_headphone_disconnect
#define sys_sound_acknowledge_headphone_disconnect STUB_sys_sound_acknowledge_headphone_disconnect
#define sys_sound_cleanup               STUB_sys_sound_cleanup
#define sys_sound_setdata               STUB_sys_sound_setdata
#define sys_sound_setvol                STUB_sys_sound_setvol
#define sys_sound_setpan                STUB_sys_sound_setpan
#define sys_sound_setfade               STUB_sys_sound_setfade
#define sys_sound_start                 STUB_sys_sound_start
#define sys_sound_stop                  STUB_sys_sound_stop
#define sys_sound_reset                 STUB_sys_sound_reset
#define sys_sound_status                STUB_sys_sound_status

#define sys_time_init                   STUB_sys_time_init
#define sys_time_now                    STUB_sys_time_now
#define sys_time_delay                  STUB_sys_time_delay

#define sys_userdata_get_data_path      STUB_sys_userdata_get_data_path
#define sys_userdata_perform            STUB_sys_userdata_perform

#define sys_display_error               STUB_sys_display_error
#define sys_get_language                STUB_sys_get_language
#define sys_open_file                   STUB_sys_open_file
#define sys_open_url                    STUB_sys_open_url
#define sys_reset_idle_timer            STUB_sys_reset_idle_timer

#define sys_debug_get_memory_stats      STUB_sys_debug_get_memory_stats

#endif  // INCLUDE_TESTS && !IN_SYSDEP && !IN_SYSDEP_TEST

/*************************************************************************/
/*************************************************************************/

#endif  // SIL_SRC_SYSDEP_H

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
