/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/android/thread.c: Thread management for Android.  Forked
 * from ../posix/thread.c to add JVM thread handling and deal with
 * Android-specific thread priority details.
 */

#define IN_SYSDEP

#include "src/base.h"
#include "src/memory.h"
#include "src/sysdep.h"
#include "src/sysdep/android/internal.h"
#include "src/sysdep/posix/thread.h"

#include <errno.h>
#include <pthread.h>
#include <sys/resource.h>
#include <unistd.h>

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Structure to hold thread data.  A pointer to this structure is returned
 * as the thread handle. */

typedef struct SysThread SysThread;
struct SysThread {
    /* Thread handle. */
    pthread_t handle;

    /* Function to call, and its parameter. */
    int (*function)(void *param);
    void *param;

    /* Priority for this thread (will be set by the thread itself once
     * it starts). */
    int priority;

    /* Flag indicating whether the thread has terminated.  We need this
     * because pthreads has no equivalent to sys_thread_is_running(). */
    uint8_t finished;
};

/*-----------------------------------------------------------------------*/

/* Local function declarations. */

/**
 * thread_runner:  Wrapper for threads started with sys_thread_create()
 * which handles setting the "finished" flag when the thread terminates
 * and returning a pointer as pthreads expects.  On Android, this function
 * also attaches the thread to / detaches it from the Dalvik VM.
 *
 * [Parameters]
 *     param: Thread parameter (points to a SysThread structure).
 * [Return value]
 *     Thread function return value.
 */
static void *thread_runner(void *param);

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

SysThreadID sys_thread_create(
    int priority, int stack_size, int (*function)(void *), void *param)
{
    #define CHECK(call)  do {                       \
        const int error = (call);                   \
        if (UNLIKELY(error != 0)) {                 \
            DLOG("%s: %s", #call, strerror(error)); \
            return 0;                               \
        }                                           \
    } while (0)

    int policy;
    struct sched_param sched_param;
    CHECK(pthread_getschedparam(pthread_self(), &policy, &sched_param));

    errno = 0;
    int my_priority = getpriority(PRIO_PROCESS, 0);
    if (my_priority == -1 && errno != 0) {
        DLOG("getpriority(PRIO_PROCESS, 0): %s", strerror(errno));
        my_priority = 0;
    }
    int real_priority = my_priority - priority;
    const int pri_min = -20;
    const int pri_max = 19;
    if (real_priority < pri_min) {
        DLOG("%p(%p): Requested priority %d (actual %d) too high, using"
             " %d (%d)", function, param, priority, real_priority,
             priority - (pri_min - real_priority), pri_min);
        real_priority = pri_min;
    } else if (real_priority > pri_max) {
        DLOG("%p(%p): Requested priority %d (actual %d) too low, using"
             " %d (%d)", function, param, priority, real_priority,
             priority - (pri_max - real_priority), pri_max);
        real_priority = pri_max;
    }

    pthread_attr_t attr;
    CHECK(pthread_attr_init(&attr));
    CHECK(pthread_attr_setschedpolicy(&attr, policy));
    CHECK(pthread_attr_setschedparam(&attr, &sched_param));
    if (stack_size > 0) {
        int page_size = sysconf(_SC_PAGESIZE);
        ASSERT(page_size > 0, page_size = 4096);
        stack_size = align_up(stack_size, page_size);
        stack_size = lbound(stack_size, (int)PTHREAD_STACK_MIN);
#ifdef __linux__
        /* Linux doesn't let you actually access the lowest page of the
         * stack (WTF?), and we don't seem to get to use the highest page,
         * so add two extra pages to the requested stack size to ensure we
         * get the usable size we want. */
        stack_size += page_size*2;
#endif
        CHECK(pthread_attr_setstacksize(&attr, stack_size));
    }

    SysThread *thread = mem_alloc(sizeof(*thread), 0, 0);
    if (UNLIKELY(!thread)) {
        DLOG("No memory for thread structure");
        return 0;
    }
    thread->function = function;
    thread->param = param;
    thread->priority = real_priority;
    thread->finished = 0;

    const int error = pthread_create(&thread->handle, &attr,
                                     thread_runner, thread);
    pthread_attr_destroy(&attr);
    if (error != 0) {
        DLOG("Failed to create thread for %p(%p): %s", function, param,
             strerror(error));
        mem_free(thread);
        return 0;
    }

    return (SysThreadID)thread;

    #undef CHECK
}

/*-----------------------------------------------------------------------*/

int sys_thread_is_running(SysThreadID thread_)
{
    SysThread *thread = (SysThread *)thread_;
    BARRIER();
    return !thread->finished;
}

/*-----------------------------------------------------------------------*/

int sys_thread_wait(SysThreadID thread_, int *result_ret)
{
    SysThread *thread = (SysThread *)thread_;

    /* Some older versions of Android fail to detect waiting on self. */
    if (UNLIKELY(thread->handle == pthread_self())) {
        DLOG("pthread_join(%p): %s", &thread->handle, strerror(EDEADLK));
        return 0;
    }

    void *retval = NULL;
    int error = pthread_join(thread->handle, &retval);
    if (error != 0) {
        DLOG("pthread_join(%p): %s", &thread->handle, strerror(error));
        return 0;
    }
    mem_free(thread);
    /* Convert through intptr_t to silence compiler warnings about
     * converting from a pointer to a differently-sized integer. */
    *result_ret = (int)(intptr_t)retval;
    return 1;
}

/*-----------------------------------------------------------------------*/

void sys_thread_yield()
{
    BARRIER();
    sched_yield();
    BARRIER();
}

/*************************************************************************/
/*********************** Exported utility routines ***********************/
/*************************************************************************/

int posix_thread_create_detached(void (*function)(void *), void *param)
{
    #define CHECK(call)  do {                       \
        const int error = (call);                   \
        if (UNLIKELY(error != 0)) {                 \
            DLOG("%s: %s", #call, strerror(error)); \
            return 0;                               \
        }                                           \
    } while (0)

    pthread_attr_t attr;
    CHECK(pthread_attr_init(&attr));
    CHECK(pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED));

    pthread_t handle;
    const int error = pthread_create(&handle, &attr,
                                     (void *(*)(void *))function, param);
    pthread_attr_destroy(&attr);
    if (error == 0) {
        return 1;
    } else {
        DLOG("Failed to create thread for %p(%p): %s", function, param,
             strerror(error));
        return 0;
    }

    #undef CHECK
}

/*-----------------------------------------------------------------------*/

void posix_thread_exit(int retval)
{
    JavaVM *vm = android_activity->vm;
    (*vm)->DetachCurrentThread(vm);
    pthread_exit((void *)(intptr_t)retval);
}

/*************************************************************************/
/**************************** Local routines *****************************/
/*************************************************************************/

static void *thread_runner(void *param)
{
    SysThread *thread = (SysThread *)param;

    JavaVM *vm = android_activity->vm;
    JNIEnv *env;
    (*vm)->AttachCurrentThread(vm, &env, NULL);

    jclass android_os_Process = get_class("android.os.Process");
    ASSERT(android_os_Process, return NULL);
    jmethodID setThreadPriority = (*env)->GetStaticMethodID(
        env, android_os_Process, "setThreadPriority", "(I)V");
    ASSERT(setThreadPriority != 0, return NULL);
    (*env)->CallStaticVoidMethod(env, android_os_Process,
                                 setThreadPriority, thread->priority);
    if (clear_exceptions(env)) {
        DLOG("Failed to set thread priority to %d", thread->priority);
    }

    const int retval = (*thread->function)(thread->param);

    (*vm)->DetachCurrentThread(vm);

    /* Make sure all actions from the thread function happen before the
     * thread->finished store is seen. */
    BARRIER();
    thread->finished = 1;

    return (void *)(intptr_t)retval;
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
