/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/macosx/userdata.c: User data access routines for Mac OS X.
 */

/*
 * The default pathnames for this implementation are as follows:
 *
 *    - Save files: <user-data-path>/save/save-NNNN.bin
 *         (NNNN is the save number, zero-padded to 4 digits)
 *    - Settings file: <user-data-path>/settings.bin
 *    - Per-user statistics file: <user-data-path>/stats.bin
 *    - Arbitrary data files: <user-data-path>/<datafile-path>
 *
 * where <user-data-path> is the path returned by userdata_get_data_path()
 * minus the trailing slash.
 *
 * userdata_get_data_path() is supported, and returns
 * "$HOME/Library/Application Support/<program-name>/".
 *
 * "$HOME" in the above pathnames is replaced by the user's home directory
 * as found in the environment variable $HOME, or "." if that variable is
 * missing or empty.
 *
 * See ../posix/userdata.c for further details.
 */

#define IN_SYSDEP

#include "src/base.h"
#include "src/memory.h"
#include "src/sysdep.h"

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

char *sys_userdata_get_data_path(const char *program_name)
{
    PRECOND(program_name != NULL, return NULL);

    const char *home = getenv("HOME");
    if (!home || !*home) {
        home = ".";  // Fall back to the current directory if necessary.
    }

    const int path_size = strlen(home)       // $HOME
                        + 29                 // "/Library/Application Support/"
                        + strlen(program_name) // program_name
                        + 2;                 // "/\0"
    char *path = mem_alloc(path_size, 0, 0);
    if (UNLIKELY(!path)) {
        DLOG("Out of memory generating user data directory path (%u bytes)",
             path_size);
        return NULL;
    }

    const int path_len = strformat(path, path_size,
                                   "%s/Library/Application Support/%s/",
                                   home, program_name);
    if (UNLIKELY(path_len != path_size-1)) {
        DLOG("Internal error: path_len(%u) != path_size(%u) - 1",
             path_len, path_size);
        mem_free(path);
        return NULL;
    }

    return path;
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
