/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/test/sysdep/opengl/framebuffer.c: OpenGL-specific framebuffer tests.
 */

#ifdef INCLUDE_TESTS  // To the end of the file.

#include "src/base.h"
#include "src/graphics.h"
#include "src/sysdep.h"
#include "src/sysdep/opengl/internal.h"
#include "src/sysdep/opengl/opengl.h"
#include "src/test/base.h"
#include "src/test/graphics/internal.h"
#include "src/texture.h"

/*************************************************************************/
/****************************** Test runner ******************************/
/*************************************************************************/

static int wrap_test_opengl_framebuffer(void);
static int do_test_opengl_framebuffer(void);
int test_opengl_framebuffer(void)
{
    return run_tests_in_window(wrap_test_opengl_framebuffer);
}

static int wrap_test_opengl_framebuffer(void)
{
    if (!opengl_has_features(OPENGL_FEATURE_FRAMEBUFFERS)) {
        SKIP("Framebuffers not available.");
        /* We can still test the framebuffers-disabled path. */
        return test_no_framebuffer_support();
    }
    return do_test_opengl_framebuffer();
}

DEFINE_GENERIC_TEST_RUNNER(do_test_opengl_framebuffer)

TEST_INIT(init)
{
    graphics_start_frame();
    return 1;
}

TEST_CLEANUP(cleanup)
{
    graphics_finish_frame();
    opengl_free_dead_resources(1);
    return 1;
}

/*************************************************************************/
/*************************************************************************/

TEST(test_no_framebuffer_support)
{
    graphics_clear(0, 0, 0);
    graphics_finish_frame();
    graphics_cleanup();
    TEST_opengl_force_feature_mask = OPENGL_FEATURE_FRAMEBUFFERS;
    TEST_opengl_force_feature_flags = 0;
    ASSERT(graphics_init());
    ASSERT(open_window(TESTW, TESTH));
    graphics_start_frame();
    TEST_opengl_force_feature_mask = 0;
    TEST_opengl_force_feature_flags = 0;

    CHECK_FALSE(sys_framebuffer_create(16, 16));

    graphics_clear(0, 0, 0);
    graphics_finish_frame();
    graphics_cleanup();
    ASSERT(graphics_init());
    ASSERT(open_window(TESTW, TESTH));
    graphics_start_frame();
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_create_failure)
{
    CHECK_FALSE(sys_framebuffer_create(0, 0));
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_restore_all)
{
    SysFramebuffer *framebuffer1, *framebuffer2;
    CHECK_TRUE(framebuffer1 = sys_framebuffer_create(16, 16));
    CHECK_TRUE(framebuffer2 = sys_framebuffer_create(32, 32));

    /* Manually delete the GL objects to simulate device loss. */
    glDeleteFramebuffers(1, &framebuffer1->framebuffer);
    glDeleteFramebuffers(1, &framebuffer2->framebuffer);

    opengl_framebuffer_clear_all();
    opengl_framebuffer_restore_all();

    /* Check that the framebuffers are recreated and usable. */

    SysTexture *texture;
    uint8_t *pixels;

    sys_framebuffer_bind(framebuffer1);
    sys_graphics_clear(1, 0, 0);
    CHECK_TRUE(texture = sys_texture_create(16, 16, TEX_FORMAT_RGBA8888, 0,
                                            NULL, 0, NULL, NULL, 0, 0, 0));
    CHECK_TRUE(sys_texture_grab(texture, 0, 0, 16, 16, 1));
    CHECK_TRUE(pixels = sys_texture_lock(texture, SYS_TEXTURE_LOCK_NORMAL,
                                         0, 0, 16, 16));
    for (int i = 0; i < 16*16; i++) {
        CHECK_PIXEL(&pixels[i*4], 255,0,0,255, i/16,i%16);
    }
    sys_texture_destroy(texture);

    sys_framebuffer_bind(framebuffer1);
    sys_graphics_clear(1, 0, 0);
    CHECK_TRUE(texture = sys_texture_create(16, 16, TEX_FORMAT_RGBA8888, 0,
                                            NULL, 0, NULL, NULL, 0, 0, 0));
    CHECK_TRUE(sys_texture_grab(texture, 0, 0, 16, 16, 1));
    CHECK_TRUE(pixels = sys_texture_lock(texture, SYS_TEXTURE_LOCK_NORMAL,
                                         0, 0, 16, 16));
    for (int i = 0; i < 16*16; i++) {
        CHECK_PIXEL(&pixels[i*4], 255,0,0,255, i/16,i%16);
    }
    sys_texture_destroy(texture);

    sys_framebuffer_bind(NULL);
    sys_framebuffer_destroy(framebuffer1);
    sys_framebuffer_destroy(framebuffer2);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_bind_and_get)
{
    SysFramebuffer *framebuffer;
    CHECK_TRUE(framebuffer = sys_framebuffer_create(16, 16));
    sys_framebuffer_bind(framebuffer);

    GLint binding;
    glGetIntegerv(GL_FRAMEBUFFER_BINDING, &binding);
    CHECK_INTEQUAL(binding, framebuffer->framebuffer);

    CHECK_PTREQUAL(opengl_current_framebuffer(), framebuffer);

    sys_framebuffer_bind(NULL);
    sys_framebuffer_destroy(framebuffer);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_destroy_unbinds)
{
    SysFramebuffer *framebuffer;
    CHECK_TRUE(framebuffer = sys_framebuffer_create(16, 16));

    GLint binding;
    sys_framebuffer_bind(framebuffer);
    glGetIntegerv(GL_FRAMEBUFFER_BINDING, &binding);
    CHECK_INTEQUAL(binding, framebuffer->framebuffer);

    sys_framebuffer_destroy(framebuffer);
    opengl_free_dead_resources(0);
    glGetIntegerv(GL_FRAMEBUFFER_BINDING, &binding);
    CHECK_INTEQUAL(binding, opengl_get_default_framebuffer());

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_default_framebuffer)
{
    const GLuint saved_default_framebuffer = opengl_get_default_framebuffer();

    SysFramebuffer *framebuffer1, *framebuffer2;
    CHECK_TRUE(framebuffer1 = sys_framebuffer_create(16, 16));
    CHECK_TRUE(framebuffer2 = sys_framebuffer_create(32, 32));

    GLint binding;
    sys_framebuffer_bind(framebuffer1);
    glGetIntegerv(GL_FRAMEBUFFER_BINDING, &binding);
    CHECK_INTEQUAL(binding, framebuffer1->framebuffer);

    opengl_set_default_framebuffer(framebuffer2->framebuffer);
    sys_framebuffer_bind(NULL);
    glGetIntegerv(GL_FRAMEBUFFER_BINDING, &binding);
    /* Make sure we restore the default even if we're about to fail out. */
    opengl_set_default_framebuffer(saved_default_framebuffer);
    CHECK_INTEQUAL(binding, framebuffer2->framebuffer);

    sys_framebuffer_destroy(framebuffer1);
    sys_framebuffer_destroy(framebuffer2);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_depth_stencil_config)
{
    static const int8_t depth_list[] = {0, 16, 24, 32};
    static const int8_t stencil_list[] = {0, 8};

    for (int i = 0; i < lenof(depth_list); i++) {
        const int depth_bits = depth_list[i];
        opengl_set_depth_bits(depth_bits);
        for (int j = 0; j < lenof(stencil_list); j++) {
            const int stencil_bits = stencil_list[j];
            opengl_set_stencil_bits(stencil_bits);
            SysFramebuffer *framebuffer = sys_framebuffer_create(16, 16);
            if (!framebuffer) {
                if (depth_bits <= 16 && stencil_bits <= 8) {
                    FAIL("sys_framebuffer_create() failed for depth=%d"
                         " stencil=%d", depth_bits, stencil_bits);
                } else {
                    continue;  // Let it slide.
                }
            }
            sys_framebuffer_bind(framebuffer);
            graphics_clear(0, 0, 0);
            graphics_set_viewport(0, 0, 16, 16);
            graphics_enable_depth_test(depth_bits > 0);
            graphics_set_depth_test_comparison(GRAPHICS_COMPARISON_LESS);
            graphics_enable_stencil_test(stencil_bits > 0);
            graphics_set_stencil_comparison(GRAPHICS_COMPARISON_TRUE, 0, ~0U);
            graphics_set_stencil_operations(GRAPHICS_STENCIL_REPLACE,
                                            GRAPHICS_STENCIL_REPLACE,
                                            GRAPHICS_STENCIL_REPLACE);

            Vector4f color;

            color = (Vector4f){0,0,1,1};
            graphics_set_fixed_color(&color);
            graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS);
            graphics_add_vertex(&(Vector3f){-1,-1,0}, NULL, NULL);
            graphics_add_vertex(&(Vector3f){-1,+1,0}, NULL, NULL);
            graphics_add_vertex(&(Vector3f){+1,+1,0}, NULL, NULL);
            graphics_add_vertex(&(Vector3f){+1,-1,0}, NULL, NULL);
            graphics_end_and_draw_primitive();

            if (depth_bits > 0) {
                /* We can't properly test a 32-bit depth buffer because
                 * all the floating point stages of our rendering pipeline
                 * use single precision, causing a 2^-32 delta to be lost
                 * along the way.  The best we can do is check that it has
                 * better precision than a 24-bit depth buffer. */
                const float z = -powf(2, 2 - ubound(depth_bits,26));

                color = (Vector4f){0,1,0,1};
                graphics_set_fixed_color(&color);
                graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS);
                graphics_add_vertex(&(Vector3f){-1,-1,z}, NULL, NULL);
                graphics_add_vertex(&(Vector3f){-1,+1,z}, NULL, NULL);
                graphics_add_vertex(&(Vector3f){+1,+1,z}, NULL, NULL);
                graphics_add_vertex(&(Vector3f){+1,-1,z}, NULL, NULL);
                graphics_end_and_draw_primitive();

                /* Double-check that depth testing is actually functioning. */
                graphics_set_fixed_color(&(Vector4f){1,0,0,1});
                graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS);
                graphics_add_vertex(&(Vector3f){-1,-1,z}, NULL, NULL);
                graphics_add_vertex(&(Vector3f){-1,+1,z}, NULL, NULL);
                graphics_add_vertex(&(Vector3f){+1,+1,z}, NULL, NULL);
                graphics_add_vertex(&(Vector3f){+1,-1,z}, NULL, NULL);
                graphics_end_and_draw_primitive();
            }

            if (stencil_bits > 0) {
                const unsigned int value = 1U << (stencil_bits-1);
                const unsigned int mask = value | (value-1);

                graphics_enable_depth_test(0);
                graphics_set_stencil_comparison(GRAPHICS_COMPARISON_TRUE,
                                                value, mask);
                color = (Vector4f){0,1,1,1};
                graphics_set_fixed_color(&color);
                graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS);
                graphics_add_vertex(&(Vector3f){-1,-1,0}, NULL, NULL);
                graphics_add_vertex(&(Vector3f){-1,+1,0}, NULL, NULL);
                graphics_add_vertex(&(Vector3f){+1,+1,0}, NULL, NULL);
                graphics_add_vertex(&(Vector3f){+1,-1,0}, NULL, NULL);
                graphics_end_and_draw_primitive();

                graphics_set_stencil_comparison(GRAPHICS_COMPARISON_LESS,
                                                value-1, mask);
                graphics_set_fixed_color(&(Vector4f){1,0,1,1});
                graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS);
                graphics_add_vertex(&(Vector3f){-1,-1,0}, NULL, NULL);
                graphics_add_vertex(&(Vector3f){-1,+1,0}, NULL, NULL);
                graphics_add_vertex(&(Vector3f){+1,+1,0}, NULL, NULL);
                graphics_add_vertex(&(Vector3f){+1,-1,0}, NULL, NULL);
                graphics_end_and_draw_primitive();
            }

            const int expect_r = iroundf(color.x*255);
            const int expect_g = iroundf(color.y*255);
            const int expect_b = iroundf(color.z*255);
            uint8_t pixels[16][16][4];
            graphics_read_pixels(0, 0, 16, 16, pixels);
            for (int y = 0; y < 16; y++) {
                for (int x = 0; x < 16; x++) {
                    if (pixels[y][x][0] != expect_r
                     || pixels[y][x][1] != expect_g
                     || pixels[y][x][2] != expect_b
                     || pixels[y][x][3] != 255) {
                        FAIL("Pixel (%d,%d) was RGBA (%d,%d,%d,%d) but"
                             " should have been (%d,%d,%d,255) for depth=%d"
                             " stencil=%d", x, y, pixels[y][x][0],
                             pixels[y][x][1], pixels[y][x][2], pixels[y][x][3],
                             expect_r, expect_g, expect_b, depth_bits,
                             stencil_bits);
                    }
                    CHECK_PIXEL(pixels[y][x],
                                expect_r, expect_g, expect_b, 255, x, y);
                }
            }

            sys_framebuffer_destroy(framebuffer);
        }
    }

    graphics_enable_depth_test(0);
    graphics_enable_stencil_test(0);
    opengl_set_depth_bits(16);
    opengl_set_stencil_bits(8);
    return 1;
}

/*************************************************************************/
/*************************************************************************/

#endif  // INCLUDE_TESTS

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
